@@ -38,8 +38,9 @@ GetOptions(\%args, qw(
     ar=s cc=s ld=s make=s has-sha has-libuv
     static has-libtommath has-libatomic_ops
     has-dyncall has-libffi pkgconfig=s
-    build=s host=s big-endian jit! enable-jit lua=s has-dynasm
+    build=s host=s big-endian jit! enable-jit
     prefix=s bindir=s libdir=s mastdir=s make-install asan ubsan valgrind telemeh),
+
     'no-optimize|nooptimize' => sub { $args{optimize} = 0 },
     'no-debug|nodebug' => sub { $args{debug} = 0 },
     'no-telemeh|notelemeh' => sub { $args{telemeh} = 0 }
@@ -77,11 +78,13 @@ if (-d '.git') {
 $args{optimize}     = 3 if not defined $args{optimize} or $args{optimize} eq "";
 $args{debug}        = 3 if defined $args{debug} and $args{debug} eq "";
 
+
 for (qw(coverage instrument static big-endian has-libtommath has-sha has-libuv
-        has-libatomic_ops has-dynasm asan ubsan valgrind)) {
+        has-libatomic_ops asan ubsan valgrind)) {
     $args{$_} = 0 unless defined $args{$_};
 }
 
+
 # jit is default
 $args{jit} = 1 unless defined $args{jit};
 
@@ -98,9 +101,10 @@ $config{perl}   = $^X;
 $config{config} = join ' ', map { / / ? "\"$_\"" : $_ } @args;
 $config{osname} = $^O;
 $config{osvers} = $Config{osvers};
-$config{lua}       = defined_or $args{lua}, './3rdparty/dynasm/minilua@exe@';
 $config{pkgconfig} = defined_or $args{pkgconfig}, '/usr/bin/pkg-config';
 
+
+
 # set options that take priority over all others
 my @keys = qw( ar cc ld make );
 @config{@keys} = @args{@keys};
@@ -143,11 +147,25 @@ $config{translate_newline_output} = 0  unless defined $config{translate_newline_
 $config{ld}           = $config{cc} unless defined $config{ld};
 $config{ldout}        = $config{ccout} unless defined $config{ldout};
 $config{ldsys}        = $config{ldusr} unless defined $config{ldsys};
-$config{ldmiscflags}  = $config{ccmiscflags} unless defined $config{ldmiscflags};
 $config{ldoptiflags}  = $config{ccoptiflags} unless defined $config{ldoptiflags};
 $config{lddebugflags} = $config{ccdebugflags} unless defined $config{lddebugflags};
 $config{ldinstflags}  = $config{ccinstflags} unless defined $config{ldinstflags};
 
+# Probe the compiler.
+build::probe::compiler_usability(\%config, \%defaults);
+
+# Remove unsupported -Werror=* gcc flags if gcc doesn't support them.
+build::probe::specific_werror(\%config, \%defaults);
+if ($config{cc} eq 'gcc' && !$config{can_specific_werror}) {
+    $config{ccmiscflags} =~ s/-Werror=[^ ]+//g;
+    $config{ccmiscflags} =~ s/ +/ /g;
+    $config{ccmiscflags} =~ s/^ +$//;
+}
+
+# Set the remaining ldmiscflags. Do this after probing for gcc -Werror probe to not miss that change for the linker.
+$config{ldmiscflags}  = $config{ccmiscflags} unless defined $config{ldmiscflags};
+
+
 if ($args{'has-sha'}) {
     $config{shaincludedir} = '/usr/include/sha';
     $defaults{-thirdparty}->{sha} = undef;
@@ -167,7 +185,7 @@ if (-e '3rdparty/libuv/src/unix/threadpool' . $defaults{obj}
 
 # test whether pkg-config works
 if (-e "$config{pkgconfig}") {
-    print("\nTesting pkgconfig ... ");
+    print dots("    Testing pkgconfig");
     system("$config{pkgconfig}", "--version");
     if ( $? == 0 ) {
         $config{pkgconfig_works} = 1;
@@ -248,15 +266,6 @@ else {
                         . "\t\$(CP) 3rdparty/libtommath/*.h \$(DESTDIR)\$(PREFIX)/include/libtommath\n";
 }
 
-if ($args{'has-dynasm'}) {
-    $config{dynasmlua}  = '-l dynasm.lua';
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '/usr/include/luajit-2.0';
-}
-else {
-    $config{dynasmlua}  = './3rdparty/dynasm/dynasm.lua';
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/dynasm';
-}
-
 if ($args{'has-libffi'}) {
     $config{nativecall_backend} = 'libffi';
     unshift @{$config{usrlibs}}, 'ffi';
@@ -317,8 +326,8 @@ else {
 $config{ldlibs} = join ' ',
     (map { sprintf $config{ldusr}, $_; } @{$config{usrlibs}}),
     (map { sprintf $config{ldsys}, $_; } @{$config{syslibs}});
-$config{ldlibs} = ' -lasan ' . $config{ldlibs} if $args{asan};
-$config{ldlibs} = ' -lubsan ' . $config{ldlibs} if $args{ubsan};
+$config{ldlibs} = ' -lasan ' . $config{ldlibs} if $args{asan} and $^O ne 'darwin';
+$config{ldlibs} = ' -lubsan ' . $config{ldlibs} if $args{ubsan} and $^O ne 'darwin';
 # macro defs
 $config{ccdefflags} = join ' ', map { $config{ccdef} . $_ } @{$config{defs}};
 
@@ -341,6 +350,7 @@ push @cflags, $config{ccshared}     unless $args{static};
 push @cflags, '-fno-omit-frame-pointer' if $args{asan} or $args{ubsan};
 push @cflags, '-fsanitize=address' if $args{asan};
 push @cflags, '-fsanitize=undefined' if $args{ubsan};
+push @cflags, '-DWSL_BASH_ON_WIN' if wsl_bash_on_win();
 push @cflags, '-DDEBUG_HELPERS' if $args{debug};
 push @cflags, '-DMVM_VALGRIND_SUPPORT' if $args{valgrind};
 push @cflags, '-DHAVE_TELEMEH' if $args{telemeh};
@@ -356,7 +366,7 @@ push @ldflags, $config{lddebugflags} if $args{debug};
 push @ldflags, $config{ldinstflags}       if $args{instrument};
 push @ldflags, $config{ld_covflags}  if $args{coverage};
 push @ldflags, $config{ldrpath}           if not $args{static} and $config{prefix} ne '/usr';
-push @ldflags, $^O eq 'darwin' ? '-faddress-sanitizer' : '-fsanitize=address' if $args{asan};
+push @ldflags, '-fsanitize=address' if $args{asan};
 push @ldflags, $ENV{LDFLAGS}  if $ENV{LDFLAGS};
 $config{ldflags} = join ' ', @ldflags;
 
@@ -389,36 +399,47 @@ $config{mainlibs} = '-lubsan ' . $config{mainlibs} if $args{ubsan};
 my @auxfiles = @{ $defaults{-auxfiles} };
 $config{auxclean} = @auxfiles ? '$(RM) ' . join ' ', @auxfiles : '@:';
 
-print "OK\n";
+print "OK\n\n";
 
 if ($config{crossconf}) {
     build::auto::detect_cross(\%config, \%defaults);
-    build::probe::compiler_usability(\%config, \%defaults);
     build::probe::static_inline_cross(\%config, \%defaults);
     build::probe::unaligned_access_cross(\%config, \%defaults);
     build::probe::ptr_size_cross(\%config, \%defaults);
 }
 else {
     build::auto::detect_native(\%config, \%defaults);
-    build::probe::compiler_usability(\%config, \%defaults);
     build::probe::static_inline_native(\%config, \%defaults);
     build::probe::unaligned_access(\%config, \%defaults);
     build::probe::ptr_size_native(\%config, \%defaults);
 }
 
+
 if ($args{'jit'}) {
     if ($config{ptr_size} != 8) {
         print "JIT isn't supported on platforms with $config{ptr_size} byte pointers.\n";
     } elsif ($Config{archname} =~ m/^x86_64|^amd64|^darwin(-thread)?(-multi)?-2level/) {
-        $config{jit} = '$(JIT_POSIX_X64)';
+        $config{jit_obj}      = '$(JIT_OBJECTS) $(JIT_ARCH_X64)';
+        $config{dasm_flags}   = '-D POSIX=1';
+        $config{jit_arch}     = 'MVM_JIT_ARCH_X64';
+        $config{jit_platform} = 'MVM_JIT_PLATFORM_POSIX';
     } elsif ($Config{archname} =~ /^MSWin32-x64/) {
-        $config{jit} = '$(JIT_WIN32_X64)';
+        $config{jit_obj}      = '$(JIT_OBJECTS) $(JIT_ARCH_X64)';
+        $config{dasm_flags}   = '-D WIN32=1';
+        $config{jit_arch}     = 'MVM_JIT_ARCH_X64';
+        $config{jit_platform} = 'MVM_JIT_PLATFORM_WIN32';
     } else {
         print "JIT isn't supported on $Config{archname} yet.\n";
     }
 }
 # fallback
-$config{jit} = '$(JIT_STUB)' unless defined $config{jit};
+unless (defined $config{jit_obj}) {
+    $config{jit_obj}      = '$(JIT_STUB)';
+    $config{jit_arch}     = 'MVM_JIT_ARCH_NONE';
+    $config{jit_platform} = 'MVM_JIT_PLATFORM_NONE';
+    $config{dasm_flags}   = '';
+}
+
 
 if ($config{cc} eq 'cl') {
     $config{install}   .= "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/msinttypes\n"
@@ -572,6 +593,7 @@ sub setup_native {
     my ($os) = @_;
 
     print dots("Configuring native build environment");
+    print "\n";
 
     $os = build::probe::win32_compiler_toolchain(\%config, \%defaults)
         if $os eq 'MSWin32';
@@ -745,8 +767,9 @@ sub generate {
 sub dots {
     my $message = shift;
     my $length = shift || 55;
-
-    return "$message ". '.' x ($length - length $message) . ' ';
+    my $dot_count = $length - length $message;
+    $dot_count = 0 if $dot_count < 0;
+    return "$message " . '.' x $dot_count . ' ';
 }
 
 # fail but continue
@@ -788,6 +811,10 @@ sub write_backend_config {
     }
 }
 
+sub wsl_bash_on_win {
+    open my $fh, '<', '/proc/sys/kernel/osrelease' or return 0;
+    return ((readline $fh) =~ /\A\d\.\d\.\d-\d+-Microsoft\s*\z/) ? 1 : 0;
+}
 
 __END__
 
@@ -801,15 +828,15 @@ __END__
                    [--debug] [--optimize] [--instrument]
                    [--static] [--prefix]
                    [--has-libtommath] [--has-sha] [--has-libuv]
-                   [--has-libatomic_ops] [--has-dynasm]
-                   [--lua <lua>] [--asan] [--ubsan] [--no-jit]
+                   [--has-libatomic_ops]
+                   [--asan] [--ubsan] [--no-jit]
                    [--telemeh]
 
     ./Configure.pl --build <build-triple> --host <host-triple>
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
                    [--debug] [--optimize] [--instrument]
                    [--static] [--big-endian] [--prefix]
-                   [--lua <lua>] [--make-install]
+                   [--make-install]
 
 =head2 Use of environment variables
 
@@ -956,8 +983,6 @@ Build and install MoarVM in addition to configuring it.
 
 =item --has-libatomic_ops
 
-=item --has-dynasm
-
 =item --has-dyncall
 
 =item --has-libffi
@@ -970,10 +995,6 @@ Provide path to the pkgconfig executable. Default: /usr/bin/pkg-config
 
 Disable JIT compiler, which is enabled by default to JIT-compile hot frames.
 
-=item --lua=path/to/lua/executable
-
-Path to a lua executable. (Used during the build when JIT is enabled).
-
 =item --telemeh
 
 Build support for the fine-grained internal event logger.
