@@ -581,6 +581,8 @@ static void string_consts(MVMThreadContext *tc) {
     string_creator(translate_newlines, "translate_newlines");
     string_creator(platform_newline, MVM_TRANSLATE_NEWLINE_OUTPUT ? "\r\n" : "\n");
     string_creator(path, "path");
+    string_creator(config, "config");
+    string_creator(replacement, "replacement");
 }
 
 /* Drives the overall bootstrap process. */
