@@ -111,6 +111,8 @@ static void at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *d
         interval_id = MVM_telemetry_interval_start(tc, "ConcBlockingQueue.at_pos");
         MVMROOT(tc, root, {
             MVM_gc_mark_thread_blocked(tc);
+            data = OBJECT_BODY(root);
+            cbq = (MVMConcBlockingQueueBody *)data;
             uv_mutex_lock(&cbq->locks->head_lock);
             MVM_gc_mark_thread_unblocked(tc);
             data = OBJECT_BODY(root);
@@ -128,7 +130,7 @@ static void at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *d
 
 static MVMuint64 elems(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data) {
     MVMConcBlockingQueueBody *cbq = (MVMConcBlockingQueueBody *)data;
-    return MVM_load(cbq->elems);
+    return MVM_load(&(cbq->elems));
 }
 
 static void push(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMRegister value, MVMuint16 kind) {
@@ -148,15 +150,15 @@ static void push(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *dat
     add = MVM_calloc(1, sizeof(MVMConcBlockingQueueNode));
 
     interval_id = MVM_telemetry_interval_start(tc, "ConcBlockingQueue.push");
-    MVMROOT(tc, root, {
-    MVMROOT(tc, to_add, {
+    MVMROOT2(tc, root, to_add, {
         MVM_gc_mark_thread_blocked(tc);
+        data = OBJECT_BODY(root);
+        cbq = (MVMConcBlockingQueueBody *)data;
         uv_mutex_lock(&cbq->locks->tail_lock);
         MVM_gc_mark_thread_unblocked(tc);
+        data = OBJECT_BODY(root);
+        cbq = (MVMConcBlockingQueueBody *)data;
     });
-    });
-    data = OBJECT_BODY(root);
-    cbq = (MVMConcBlockingQueueBody *)data;
     MVM_ASSIGN_REF(tc, &(root->header), add->value, to_add);
     cbq->tail->next = add;
     cbq->tail = add;
@@ -166,11 +168,13 @@ static void push(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *dat
     if (orig_elems == 0) {
         MVMROOT(tc, root, {
             MVM_gc_mark_thread_blocked(tc);
+            data = OBJECT_BODY(root);
+            cbq = (MVMConcBlockingQueueBody *)data;
             uv_mutex_lock(&cbq->locks->head_lock);
             MVM_gc_mark_thread_unblocked(tc);
+            data = OBJECT_BODY(root);
+            cbq = (MVMConcBlockingQueueBody *)data;
         });
-        data = OBJECT_BODY(root);
-        cbq = (MVMConcBlockingQueueBody *)data;
         uv_cond_signal(&cbq->locks->head_cond);
         uv_mutex_unlock(&cbq->locks->head_lock);
     }
@@ -188,6 +192,8 @@ static void shift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
     interval_id = MVM_telemetry_interval_start(tc, "ConcBlockingQueue.shift");
     MVMROOT(tc, root, {
         MVM_gc_mark_thread_blocked(tc);
+        data = OBJECT_BODY(root);
+        cbq = (MVMConcBlockingQueueBody *)data;
         uv_mutex_lock(&cbq->locks->head_lock);
         MVM_gc_mark_thread_unblocked(tc);
         data = OBJECT_BODY(root);
@@ -195,6 +201,8 @@ static void shift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
 
         while (MVM_load(&cbq->elems) == 0) {
                 MVM_gc_mark_thread_blocked(tc);
+                data = OBJECT_BODY(root);
+                cbq = (MVMConcBlockingQueueBody *)data;
                 uv_cond_wait(&cbq->locks->head_cond, &cbq->locks->head_lock);
                 MVM_gc_mark_thread_unblocked(tc);
                 data = OBJECT_BODY(root);
@@ -247,7 +255,9 @@ static const MVMREPROps ConcBlockingQueue_this_repr = {
         MVM_REPR_DEFAULT_BIND_POS_MULTIDIM,
         MVM_REPR_DEFAULT_DIMENSIONS,
         MVM_REPR_DEFAULT_SET_DIMENSIONS,
-        MVM_REPR_DEFAULT_GET_ELEM_STORAGE_SPEC
+        MVM_REPR_DEFAULT_GET_ELEM_STORAGE_SPEC,
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC,
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM
     },    /* pos_funcs */
     MVM_REPR_DEFAULT_ASS_FUNCS,
     elems,
