@@ -169,7 +169,9 @@ static const MVMREPROps MVMIter_this_repr = {
         MVM_REPR_DEFAULT_BIND_POS_MULTIDIM,
         MVM_REPR_DEFAULT_DIMENSIONS,
         MVM_REPR_DEFAULT_SET_DIMENSIONS,
-        get_elem_storage_spec
+        get_elem_storage_spec,
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC,
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM
     },    /* pos_funcs */
     MVM_REPR_DEFAULT_ASS_FUNCS,
     MVM_REPR_DEFAULT_ELEMS,
@@ -195,6 +197,9 @@ static const MVMREPROps MVMIter_this_repr = {
 
 MVMObject * MVM_iter(MVMThreadContext *tc, MVMObject *target) {
     MVMIter *iterator;
+    if (!IS_CONCRETE(target)) {
+        MVM_exception_throw_adhoc(tc, "Cannot iterate over a %s type object", MVM_6model_get_debug_name(tc, target));
+    }
     MVMROOT(tc, target, {
         if (REPR(target)->ID == MVM_REPR_ID_VMArray) {
             iterator = (MVMIter *)MVM_repr_alloc_init(tc,
@@ -232,8 +237,7 @@ MVMObject * MVM_iter(MVMThreadContext *tc, MVMObject *target) {
                 MVMStaticFrame      *sf     = frame->static_info;
                 MVMLexicalRegistry **lexreg = sf->body.lexical_names_list;
                 MVMuint32 i;
-                MVMROOT(tc, frame, {
-                MVMROOT(tc, sf, {
+                MVMROOT2(tc, frame, sf, {
                     for (i = 0; i < sf->body.num_lexicals; i++) {
                         MVMuint32 idx  = lexreg[i]->value;
                         MVMuint16 type = sf->body.lexical_types[idx];
@@ -257,24 +261,48 @@ MVMObject * MVM_iter(MVMThreadContext *tc, MVMObject *target) {
                                 MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
                                 break;
                             }
+                            case MVM_reg_uint8: {
+                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
+                                    frame->env[idx].u8);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                                break;
+                            }
                             case MVM_reg_int16: {
                                 MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                                     frame->env[idx].i16);
                                 MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
                                 break;
                             }
+                            case MVM_reg_uint16: {
+                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
+                                    frame->env[idx].u16);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                                break;
+                            }
                             case MVM_reg_int32: {
                                 MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                                     frame->env[idx].i32);
                                 MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
                                 break;
                             }
+                            case MVM_reg_uint32: {
+                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
+                                    frame->env[idx].u32);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                                break;
+                            }
                             case MVM_reg_int64: {
                                 MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                                     frame->env[idx].i64);
                                 MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
                                 break;
                             }
+                            case MVM_reg_uint64: {
+                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
+                                    frame->env[idx].u64);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                                break;
+                            }
                             case MVM_reg_num32: {
                                 MVMObject *bn = MVM_repr_box_num(tc, hll->num_box_type,
                                     frame->env[idx].n32);
@@ -289,11 +317,10 @@ MVMObject * MVM_iter(MVMThreadContext *tc, MVMObject *target) {
                             }
                             default:
                                 MVM_exception_throw_adhoc(tc,
-                                    "Unknown lexical type encountered while building context iterator");
+                                    "%s lexical type encountered while building context iterator", MVM_reg_get_debug_name(tc, type));
                         }
                     }
                 });
-                });
             });
 
             /* Call ourselves recursively to get the iterator for this
@@ -302,7 +329,7 @@ MVMObject * MVM_iter(MVMThreadContext *tc, MVMObject *target) {
         }
         else {
             MVM_exception_throw_adhoc(tc, "Cannot iterate object with %s representation (%s)",
-                REPR(target)->name, STABLE(target)->debug_name);
+                REPR(target)->name, MVM_6model_get_debug_name(tc, target));
         }
     });
     return (MVMObject *)iterator;
@@ -327,7 +354,7 @@ MVMint64 MVM_iter_istrue(MVMThreadContext *tc, MVMIter *iter) {
 MVMString * MVM_iterkey_s(MVMThreadContext *tc, MVMIter *iterator) {
     if (REPR(iterator)->ID != MVM_REPR_ID_MVMIter
             || iterator->body.mode != MVM_ITER_MODE_HASH)
-        MVM_exception_throw_adhoc(tc, "This is not a hash iterator, it's a %s (%s)", REPR(iterator)->name, STABLE(iterator)->debug_name);
+        MVM_exception_throw_adhoc(tc, "This is not a hash iterator, it's a %s (%s)", REPR(iterator)->name, MVM_6model_get_debug_name(tc, (MVMObject *)iterator));
     if (!iterator->body.hash_state.curr)
         MVM_exception_throw_adhoc(tc, "You have not advanced to the first item of the hash iterator, or have gone past the end");
     return MVM_HASH_KEY(iterator->body.hash_state.curr);
@@ -338,7 +365,7 @@ MVMObject * MVM_iterval(MVMThreadContext *tc, MVMIter *iterator) {
     MVMObject *target;
     MVMRegister result;
     if (REPR(iterator)->ID != MVM_REPR_ID_MVMIter)
-        MVM_exception_throw_adhoc(tc, "This is not an iterator, it's a %s (%s)", REPR(iterator)->name, STABLE(iterator)->debug_name);
+        MVM_exception_throw_adhoc(tc, "This is not an iterator, it's a %s (%s)", REPR(iterator)->name, MVM_6model_get_debug_name(tc, (MVMObject *)iterator));
     if (iterator->body.mode == MVM_ITER_MODE_ARRAY) {
         body = &iterator->body;
         if (body->array_state.index == -1)
