@@ -47,9 +47,9 @@ static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     MVMint64 i;
     for (i = 0; i < nfa->body.num_states; i++)
         if (nfa->body.num_state_edges[i])
-            MVM_free(nfa->body.states[i]);
-    MVM_free(nfa->body.states);
-    MVM_free(nfa->body.num_state_edges);
+            MVM_fixed_size_free(tc, tc->instance->fsa, nfa->body.num_state_edges[i] * sizeof(MVMNFAStateInfo), nfa->body.states[i]);
+    MVM_fixed_size_free(tc, tc->instance->fsa, nfa->body.num_states * sizeof(MVMNFAStateInfo *), nfa->body.states);
+    MVM_fixed_size_free(tc, tc->instance->fsa, nfa->body.num_states * sizeof(MVMint64), nfa->body.num_state_edges);
 }
 
 
@@ -107,10 +107,9 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
                          * and then each of the codepoints. */
                         MVMNFGSynthetic *si = MVM_nfg_get_synthetic_info(tc, g);
                         MVMint32 k;
-                        MVM_serialization_write_int(tc, writer, -(si->num_combs + 1));
-                        MVM_serialization_write_int(tc, writer, si->base);
-                        for (k = 0; k < si->num_combs; k++)
-                            MVM_serialization_write_int(tc, writer, si->combs[k]);
+                        MVM_serialization_write_int(tc, writer, -(si->num_codes));
+                        for (k = 0; k < si->num_codes; k++)
+                            MVM_serialization_write_int(tc, writer, si->codes[k]);
                     }
                     break;
                 }
@@ -153,16 +152,17 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
 
     if (body->num_states > 0) {
         /* Read state edge list counts. */
-        body->num_state_edges = MVM_malloc(body->num_states * sizeof(MVMint64));
+        body->num_state_edges = MVM_fixed_size_alloc(tc, tc->instance->fsa, body->num_states * sizeof(MVMint64));
         for (i = 0; i < body->num_states; i++)
             body->num_state_edges[i] = MVM_serialization_read_int(tc, reader);
 
         /* Read state graph. */
-        body->states = MVM_malloc(body->num_states * sizeof(MVMNFAStateInfo *));
+        body->states = MVM_fixed_size_alloc(tc, tc->instance->fsa, body->num_states * sizeof(MVMNFAStateInfo *));
         for (i = 0; i < body->num_states; i++) {
             MVMint64 edges = body->num_state_edges[i];
-            if (edges > 0)
-                body->states[i] = MVM_malloc(edges * sizeof(MVMNFAStateInfo));
+            if (edges > 0) {
+                body->states[i] = MVM_fixed_size_alloc(tc, tc->instance->fsa, edges * sizeof(MVMNFAStateInfo));
+            }
             for (j = 0; j < edges; j++) {
                 body->states[i][j].act = MVM_serialization_read_int(tc, reader);
                 body->states[i][j].to = MVM_serialization_read_int(tc, reader);
@@ -181,12 +181,12 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
                         }
                         else {
                             MVMint32 num_codes = -cp_or_synth_count;
-                            MVMCodepoint *codes = MVM_malloc(num_codes * sizeof(MVMCodepoint));
+                            MVMCodepoint *codes = MVM_fixed_size_alloc(tc, tc->instance->fsa, num_codes * sizeof(MVMCodepoint));
                             MVMint32 k;
                             for (k = 0; k < num_codes; k++)
                                 codes[k] = (MVMCodepoint)MVM_serialization_read_int(tc, reader);
                             body->states[i][j].arg.g = MVM_nfg_codes_to_grapheme(tc, codes, num_codes);
-                            MVM_free(codes);
+                            MVM_fixed_size_free(tc, tc->instance->fsa, num_codes * sizeof(MVMCodepoint), codes);
                         }
                         break;
                     }
@@ -305,8 +305,7 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
     MVMNFABody *nfa;
     MVMint64    i, j, num_states;
 
-    MVMROOT(tc, states, {
-    MVMROOT(tc, nfa_type, {
+    MVMROOT2(tc, states, nfa_type, {
         /* Create NFA object. */
         nfa_obj = MVM_repr_alloc_init(tc, nfa_type);
         nfa = (MVMNFABody *)OBJECT_BODY(nfa_obj);
@@ -318,8 +317,8 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
         num_states = MVM_repr_elems(tc, states) - 1;
         nfa->num_states = num_states;
         if (num_states > 0) {
-            nfa->num_state_edges = MVM_calloc(num_states, sizeof(MVMint64));
-            nfa->states = MVM_calloc(num_states, sizeof(MVMNFAStateInfo *));
+            nfa->num_state_edges = MVM_fixed_size_alloc_zeroed(tc, tc->instance->fsa, num_states * sizeof(MVMint64));
+            nfa->states = MVM_fixed_size_alloc_zeroed(tc, tc->instance->fsa, num_states * sizeof(MVMNFAStateInfo *));
         }
         for (i = 0; i < num_states; i++) {
             MVMObject *edge_info = MVM_repr_at_pos_o(tc, states, i + 1);
@@ -328,8 +327,9 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
             MVMint64   cur_edge  = 0;
 
             nfa->num_state_edges[i] = edges;
-            if (edges > 0)
-                nfa->states[i] = MVM_malloc(edges * sizeof(MVMNFAStateInfo));
+            if (edges > 0) {
+                nfa->states[i] = MVM_fixed_size_alloc(tc, tc->instance->fsa, edges * sizeof(MVMNFAStateInfo));
+            }
 
             for (j = 0; j < elems; j += 3) {
                 MVMint64 act = MVM_coerce_simple_intify(tc,
@@ -390,19 +390,27 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
             }
         }
     });
-    });
 
     return nfa_obj;
 }
 
 /* Does a run of the NFA. Produces a list of integers indicating the
  * chosen ordering. */
+static MVMint32 in_done(MVMuint32 *done, MVMuint32 numdone, MVMuint32 st) {
+    MVMuint32 i = 0;
+    for (i = 0; i < numdone; i++)
+        if (done[i] == st)
+            return 1;
+    return 0;
+}
 static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *target, MVMint64 offset, MVMint64 *total_fates_out) {
     MVMint64  eos     = MVM_string_graphs(tc, target);
-    MVMint64  gen     = 1;
+    MVMuint32 gen     = 1;
     MVMint64  numcur  = 0;
     MVMint64  numnext = 0;
-    MVMint64 *done, *fates, *curst, *nextst, *longlit;
+    MVMint64  numdone = 0;
+    MVMuint32 *done, *curst, *nextst;
+    MVMint64  *fates, *longlit;
     MVMint64  i, fate_arr_len, num_states, total_fates, prev_fates, usedlonglit;
     MVMint64  orig_offset = offset;
     int nfadeb = tc->instance->nfa_debug_enabled;
@@ -411,16 +419,15 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
      * states" arrays. */
     num_states = nfa->num_states;
     if (tc->nfa_alloc_states < num_states) {
-        size_t alloc   = (num_states + 1) * sizeof(MVMint64);
-        tc->nfa_done   = (MVMint64 *)MVM_realloc(tc->nfa_done, alloc);
-        tc->nfa_curst  = (MVMint64 *)MVM_realloc(tc->nfa_curst, alloc);
-        tc->nfa_nextst = (MVMint64 *)MVM_realloc(tc->nfa_nextst, alloc);
+        size_t alloc   = (num_states + 1) * sizeof(MVMuint32);
+        tc->nfa_done   = (MVMuint32 *)MVM_realloc(tc->nfa_done, alloc);
+        tc->nfa_curst  = (MVMuint32 *)MVM_realloc(tc->nfa_curst, alloc);
+        tc->nfa_nextst = (MVMuint32 *)MVM_realloc(tc->nfa_nextst, alloc);
         tc->nfa_alloc_states = num_states;
     }
     done   = tc->nfa_done;
     curst  = tc->nfa_curst;
     nextst = tc->nfa_nextst;
-    memset(done, 0, (num_states + 1) * sizeof(MVMint64));
 
     /* Allocate fates array. */
     fate_arr_len = 1 + MVM_repr_elems(tc, nfa->fates);
@@ -444,11 +451,12 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
     nextst[numnext++] = 1;
     while (numnext && offset <= eos) {
         /* Swap next and current */
-        MVMint64 *temp = curst;
+        MVMuint32 *temp = curst;
         curst   = nextst;
         nextst  = temp;
         numcur  = numnext;
         numnext = 0;
+        numdone = 0;
 
         /* Save how many fates we have before this position is considered. */
         prev_fates = total_fates;
@@ -468,9 +476,9 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
 
             MVMint64 st = curst[--numcur];
             if (st <= num_states) {
-                if (done[st] == gen)
+                if (in_done(done, numdone, st))
                     continue;
-                done[st] = gen;
+                done[numdone++] = st;
             }
 
             edge_info = nfa->states[st - 1];
@@ -528,7 +536,8 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                         fates[++j] = arg;
                         continue;
                     }
-                    else if (act == MVM_NFA_EDGE_EPSILON && to <= num_states && done[to] != gen) {
+                    else if (act == MVM_NFA_EDGE_EPSILON && to <= num_states &&
+                            !in_done(done, numdone, to)) {
                         if (to)
                             curst[numcur++] = to;
                         else if (nfadeb)  /* XXX should turn into a "can't happen" after rebootstrap */
