@@ -6,6 +6,9 @@ struct MVMSerializationRoot {
     /* The version of the serialization format. */
     MVMint32 version;
 
+    /* How many parameterized type intern entries we have */
+    MVMint32  num_param_interns;
+
     /* The SC we're serializing/deserializing. */
     MVMSerializationContext *sc;
 
@@ -44,9 +47,7 @@ struct MVMSerializationRoot {
     MVMint32  num_repos;
     char     *repos_table;
 
-    /* The number of parameterized type intern entries, and the data segment
-     * containing them. */
-    MVMint32  num_param_interns;
+    /* The the data segment containing them parameterized type intern entries */
     char     *param_interns_data;
 
     /* Array of strings making up the string heap we are constructing. If we
@@ -98,6 +99,9 @@ struct MVMSerializationReader {
     /* Number of static code objects. */
     MVMuint32 num_static_codes;
 
+    /* Whether we're already working on these worklists. */
+    MVMuint32 working;
+
     /* Array of contexts (num_contexts in length). */
     MVMFrame **contexts;
 
@@ -107,9 +111,6 @@ struct MVMSerializationReader {
     MVMDeserializeWorklist wl_objects;
     MVMDeserializeWorklist wl_stables;
 
-    /* Whether we're already working on these worklists. */
-    MVMuint32 working;
-
     /* The current object we're deserializing. */
     MVMObject *current_object;
 
@@ -125,9 +126,13 @@ struct MVMSerializationWriter {
     /* Serialization root data. */
     MVMSerializationRoot root;
 
-    /* The code refs and contexts lists we're working through/adding to. */
-    MVMObject  *codes_list;
-    MVMObject  *contexts_list;
+    /* The code refs we're working through/adding to. */
+    MVMObject *codes_list;
+
+    /* Frames that we're to serialize, along with memory management. */
+    MVMFrame **contexts_list;
+    MVMuint32 num_contexts;
+    MVMuint32 alloc_contexts;
 
     /* Current position in the stables, objects and contexts lists. */
     MVMint64 stables_list_pos;
