@@ -21,6 +21,7 @@
 /* Frame flags. */
 #define FRAME_FLAG_EXIT_HANDLER     1
 #define FRAME_FLAG_IS_THUNK         2
+#define FRAME_FLAG_NO_INLINE        8
 
 /* Describes the current reader state. */
 typedef struct {
@@ -516,6 +517,7 @@ static MVMStaticFrame ** deserialize_frames(MVMThreadContext *tc, MVMCompUnit *c
             MVMint16 flags = read_int16(pos, 38);
             static_frame_body->has_exit_handler = flags & FRAME_FLAG_EXIT_HANDLER;
             static_frame_body->is_thunk         = flags & FRAME_FLAG_IS_THUNK;
+            static_frame_body->no_inline        = flags & FRAME_FLAG_NO_INLINE;
         }
 
         /* Read code object SC indexes (version 4 and higher). */
@@ -582,7 +584,11 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
         return;
 
     /* Acquire the update mutex on the CompUnit. */
-    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
+    MVMROOT(tc, cu, {
+    MVMROOT(tc, sf, {
+        MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
+    });
+    });
 
     /* Ensure no other thread has done this for us in the mean time. */
     if (sf->body.fully_deserialized) {
@@ -648,7 +654,7 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
                 sf->body.handlers[j].label_reg = read_int16(pos, 0);
                 pos += 2;
             }
-            sf->body.handlers[j].inlined_and_not_lexical = 0;
+            sf->body.handlers[j].inlinee = -1;
         }
     }
 
