@@ -41,10 +41,9 @@ void MVM_coerce_istrue(MVMThreadContext *tc, MVMObject *obj, MVMRegister *res_re
                      * the register. */
                     MVM_args_setup_thunk(tc, res_reg, MVM_RETURN_INT, inv_arg_callsite);
                     tc->cur_frame->args[0].o = obj;
-                    if (flip) {
-                        tc->cur_frame->special_return      = flip_return;
-                        tc->cur_frame->special_return_data = res_reg;
-                    }
+                    if (flip)
+                        MVM_frame_special_return(tc, tc->cur_frame, flip_return, NULL,
+                            res_reg, NULL);
                     STABLE(code)->invoke(tc, code, inv_arg_callsite, tc->cur_frame->args);
                 }
                 else {
@@ -53,8 +52,7 @@ void MVM_coerce_istrue(MVMThreadContext *tc, MVMObject *obj, MVMRegister *res_re
                     data->true_addr  = true_addr;
                     data->false_addr = false_addr;
                     data->flip       = flip;
-                    tc->cur_frame->special_return      = boolify_return;
-                    tc->cur_frame->special_return_data = data;
+                    MVM_frame_special_return(tc, tc->cur_frame, boolify_return, NULL, data, NULL);
                     MVM_args_setup_thunk(tc, &data->res_reg, MVM_RETURN_INT, inv_arg_callsite);
                     tc->cur_frame->args[0].o = obj;
                     STABLE(code)->invoke(tc, code, inv_arg_callsite, tc->cur_frame->args);
@@ -151,9 +149,8 @@ static void flip_return(MVMThreadContext *tc, void *sr_data) {
 MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
     char buffer[64];
     int len;
-
     /* See if we can hit the cache. */
-    int cache = i >= 0 && i < MVM_INT_TO_STR_CACHE_SIZE;
+    int cache = 0 <= i && i < MVM_INT_TO_STR_CACHE_SIZE;
     if (cache) {
         MVMString *cached = tc->instance->int_to_str_cache[i];
         if (cached)
@@ -161,8 +158,8 @@ MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
     }
 
     /* Otherwise, need to do the work; cache it if in range. */
-    len = snprintf(buffer, 64, "%lld", (long long int)i);
-    if (len >= 0) {
+    len = snprintf(buffer, 64, "%"PRIi64"", i);
+    if (0 <= len) {
         MVMString *result = MVM_string_ascii_decode(tc, tc->instance->VMString, buffer, len);
         if (cache)
             tc->instance->int_to_str_cache[i] = result;
