@@ -20,6 +20,9 @@
 #define MVM_EX_CAT_EMIT          16384
 #define MVM_EX_CAT_DONE          32768
 
+/* Not a real category, but marks an inline boundary. */
+#define MVM_EX_INLINE_BOUNDARY   2147483648
+
 /* Ways to throw an exception. */
 #define MVM_EX_THROW_DYN         0
 #define MVM_EX_THROW_LEX         1
@@ -34,7 +37,7 @@ struct MVMFrameHandler {
     /* End offset into the frame's bytecode for the handler, exclusive. */
     MVMuint32 end_offset;
 
-    /* Category mask. */
+    /* Category mask or inline boundary indicator. */
     MVMuint32 category_mask;
 
     /* The kind of handler it is. */
@@ -50,11 +53,11 @@ struct MVMFrameHandler {
      * be able to check for its identity when handling e.g. `next LABEL`. */
     MVMuint16 label_reg;
 
-    /* Is this handler actually a clone of a caller's handler performed during
-     * inlining of something, and is that something not lexically enclosed in
-     * the thing it was inlined into? If yes,this flag will be set. We need to
-     * ignore such handlers when searching for matching lexical handlers only. */
-    MVMuint16 inlined_and_not_lexical;
+    /* The inlinee that this handler is associated with. Set to -1 for the
+     * top-level handlers of a frame. Used both to skip non-top-level
+     * handlers, but also to indicate, for a inline boundary indicator
+     * entry in the table, the inline whose handlers end at this point. */
+    MVMint16 inlinee;
 };
 
 /* An active (currently executing) exception handler. */
@@ -84,9 +87,6 @@ void MVM_exception_die(MVMThreadContext *tc, MVMString *str, MVMRegister *rr);
 void MVM_exception_throwobj(MVMThreadContext *tc, MVMuint8 mode, MVMObject *exObj, MVMRegister *resume_result);
 void MVM_exception_throwpayload(MVMThreadContext *tc, MVMuint8 mode, MVMuint32 cat, MVMObject *payload, MVMRegister *resume_result);
 void MVM_exception_resume(MVMThreadContext *tc, MVMObject *exObj);
-MVMObject * MVM_exception_newlexotic(MVMThreadContext *tc, MVMuint32 offset);
-MVMObject * MVM_exception_newlexotic_from_jit(MVMThreadContext *tc, MVMint32 label);
-void MVM_exception_gotolexotic(MVMThreadContext *tc, MVMint32 handler_idx, MVMStaticFrame *sf);
 MVM_PUBLIC MVM_NO_RETURN void MVM_panic_allocation_failed(size_t len) MVM_NO_RETURN_GCC;
 MVM_PUBLIC MVM_NO_RETURN void MVM_panic(MVMint32 exitCode, const char *messageFormat, ...) MVM_NO_RETURN_GCC MVM_FORMAT(printf, 2, 3);
 MVM_PUBLIC MVM_NO_RETURN void MVM_oops(MVMThreadContext *tc, const char *messageFormat, ...) MVM_NO_RETURN_GCC MVM_FORMAT(printf, 2, 3);
@@ -95,8 +95,12 @@ MVM_NO_RETURN void MVM_exception_throw_adhoc_va(MVMThreadContext *tc, const char
 MVM_PUBLIC MVM_NO_RETURN void MVM_exception_throw_adhoc_free(MVMThreadContext *tc, char **waste, const char *messageFormat, ...) MVM_NO_RETURN_GCC MVM_FORMAT(printf, 3, 4);
 MVM_NO_RETURN void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char **waste, const char *messageFormat, va_list args) MVM_NO_RETURN_GCC;
 MVM_PUBLIC void MVM_crash_on_error(void);
-char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame, MVMuint16 not_top);
-
+char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame, MVMuint16 not_top, MVMuint8 *throw_address);
+MVMint32 MVM_get_exception_category(MVMThreadContext *tc, MVMObject *ex);
+MVMObject * MVM_get_exception_payload(MVMThreadContext *tc, MVMObject *ex);
+void MVM_bind_exception_payload(MVMThreadContext *tc, MVMObject *ex, MVMObject *payload);
+void MVM_bind_exception_category(MVMThreadContext *tc, MVMObject *ex, MVMint32 category);
+void MVM_exception_returnafterunwind(MVMThreadContext *tc, MVMObject *ex);
 
 /* Exit codes for panic. */
 #define MVM_exitcode_NYI            12
