@@ -2,6 +2,8 @@
 #ifndef _WIN32
 #include <dlfcn.h>
 #endif
+#include <platform/threads.h>
+#include <platform/time.h>
 
 /* Grabs a NativeCall body. */
 MVMNativeCallBody * MVM_nativecall_get_nc_body(MVMThreadContext *tc, MVMObject *obj) {
@@ -21,7 +23,7 @@ static MVMint16 get_str_free_flag(MVMThreadContext *tc, MVMObject *info) {
     return MVM_NATIVECALL_ARG_FREE_STR;
 }
 
-/* Gets the flag for whether to free a string after a call or not. */
+/* Gets the flag for whether an arg is rw or not. */
 static MVMint16 get_rw_flag(MVMThreadContext *tc, MVMObject *info) {
     MVMString *flag = tc->instance->str_consts.rw;
     if (MVM_repr_exists_key(tc, info, flag)) {
@@ -31,6 +33,27 @@ static MVMint16 get_rw_flag(MVMThreadContext *tc, MVMObject *info) {
     return MVM_NATIVECALL_ARG_NO_RW;
 }
 
+/* Gets the flag for whether an arg is rw or not. */
+static MVMint16 get_refresh_flag(MVMThreadContext *tc, MVMObject *info) {
+    MVMString *typeobj_str = tc->instance->str_consts.typeobj;
+    if (MVM_repr_exists_key(tc, info, typeobj_str)) {
+        MVMObject *typeobj = MVM_repr_at_key_o(tc, info, typeobj_str);
+
+        if (REPR(typeobj)->ID == MVM_REPR_ID_MVMCArray) {
+            MVMCArrayREPRData  *repr_data = (MVMCArrayREPRData *)STABLE(typeobj)->REPR_data;
+
+            /* No need to refresh numbers. They're stored directly in the array. */
+            if (repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_NUMERIC)
+                return MVM_NATIVECALL_ARG_NO_REFRESH;
+
+            return MVM_NATIVECALL_ARG_REFRESH;
+        }
+    }
+
+    /* We don't know, so fail safe by assuming we have to refresh */
+    return MVM_NATIVECALL_ARG_REFRESH;
+}
+
 /* Takes a hash describing a type hands back an argument type code. */
 MVMint16 MVM_nativecall_get_arg_type(MVMThreadContext *tc, MVMObject *info, MVMint16 is_return) {
     MVMString *typename = MVM_repr_get_str(tc, MVM_repr_at_key_o(tc, info,
@@ -82,7 +105,7 @@ MVMint16 MVM_nativecall_get_arg_type(MVMThreadContext *tc, MVMObject *info, MVMi
     else if (strcmp(ctypename, "cpointer") == 0)
         result = MVM_NATIVECALL_ARG_CPOINTER | get_rw_flag(tc, info);
     else if (strcmp(ctypename, "carray") == 0)
-        result = MVM_NATIVECALL_ARG_CARRAY;
+        result = MVM_NATIVECALL_ARG_CARRAY | get_refresh_flag(tc, info);
     else if (strcmp(ctypename, "cunion") == 0)
         result = MVM_NATIVECALL_ARG_CUNION;
     else if (strcmp(ctypename, "vmarray") == 0)
@@ -146,7 +169,7 @@ MVMObject * MVM_nativecall_make_cstruct(MVMThreadContext *tc, MVMObject *type, v
         MVMCStructREPRData *repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
         if (REPR(type)->ID != MVM_REPR_ID_MVMCStruct)
             MVM_exception_throw_adhoc(tc,
-                "Native call expected return type with CStruct representation, but got a %s (%s)", REPR(type)->name, STABLE(type)->debug_name);
+                "Native call expected return type with CStruct representation, but got a %s (%s)", REPR(type)->name, MVM_6model_get_debug_name(tc, type));
         result = REPR(type)->allocate(tc, STABLE(type));
         ((MVMCStruct *)result)->body.cstruct = cstruct;
         if (repr_data->num_child_objs)
@@ -162,7 +185,7 @@ MVMObject * MVM_nativecall_make_cunion(MVMThreadContext *tc, MVMObject *type, vo
         MVMCUnionREPRData *repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
         if (REPR(type)->ID != MVM_REPR_ID_MVMCUnion)
             MVM_exception_throw_adhoc(tc,
-                "Native call expected return type with CUnion representation, but got a %s (%s)", REPR(type)->name, STABLE(type)->debug_name);
+                "Native call expected return type with CUnion representation, but got a %s (%s)", REPR(type)->name, MVM_6model_get_debug_name(tc, type));
         result = REPR(type)->allocate(tc, STABLE(type));
         ((MVMCUnion *)result)->body.cunion = cunion;
         if (repr_data->num_child_objs)
@@ -177,7 +200,7 @@ MVMObject * MVM_nativecall_make_cppstruct(MVMThreadContext *tc, MVMObject *type,
         MVMCPPStructREPRData *repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
         if (REPR(type)->ID != MVM_REPR_ID_MVMCPPStruct)
             MVM_exception_throw_adhoc(tc,
-                "Native call expected return type with CPPStruct representation, but got a %s (%s)", REPR(type)->name, STABLE(type)->debug_name);
+                "Native call expected return type with CPPStruct representation, but got a %s (%s)", REPR(type)->name, MVM_6model_get_debug_name(tc, type));
         result = REPR(type)->allocate(tc, STABLE(type));
         ((MVMCPPStruct *)result)->body.cppstruct = cppstruct;
         if (repr_data->num_child_objs)
@@ -192,7 +215,7 @@ MVMObject * MVM_nativecall_make_cpointer(MVMThreadContext *tc, MVMObject *type,
     if (ptr && type) {
         if (REPR(type)->ID != MVM_REPR_ID_MVMCPointer)
             MVM_exception_throw_adhoc(tc,
-                "Native call expected return type with CPointer representation, but got a %s (%s)", REPR(type)->name, STABLE(type)->debug_name);
+                "Native call expected return type with CPointer representation, but got a %s (%s)", REPR(type)->name, MVM_6model_get_debug_name(tc, type));
         result = REPR(type)->allocate(tc, STABLE(type));
         ((MVMCPointer *)result)->body.ptr = ptr;
     }
@@ -205,7 +228,7 @@ MVMObject * MVM_nativecall_make_carray(MVMThreadContext *tc, MVMObject *type, vo
     if (carray && type) {
         if (REPR(type)->ID != MVM_REPR_ID_MVMCArray)
             MVM_exception_throw_adhoc(tc,
-                "Native call expected return type with CArray representation, but got a %s (%s)", REPR(type)->name, STABLE(type)->debug_name);
+                "Native call expected return type with CArray representation, but got a %s (%s)", REPR(type)->name, MVM_6model_get_debug_name(tc, type));
         result = REPR(type)->allocate(tc, STABLE(type));
         ((MVMCArray *)result)->body.storage = carray;
     }
@@ -301,7 +324,7 @@ void * MVM_nativecall_unmarshal_cstruct(MVMThreadContext *tc, MVMObject *value)
         return ((MVMCStruct *)value)->body.cstruct;
     else
         MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CStruct representation, but got a %s (%s)", REPR(value)->name, STABLE(value)->debug_name);
+            "Native call expected return type with CStruct representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
 }
 
 void * MVM_nativecall_unmarshal_cppstruct(MVMThreadContext *tc, MVMObject *value) {
@@ -311,7 +334,7 @@ void * MVM_nativecall_unmarshal_cppstruct(MVMThreadContext *tc, MVMObject *value
         return ((MVMCPPStruct *)value)->body.cppstruct;
     else
         MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CPPStruct representation, but got a %s (%s)", REPR(value)->name, STABLE(value)->debug_name);
+            "Native call expected return type with CPPStruct representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
 }
 
 void * MVM_nativecall_unmarshal_cpointer(MVMThreadContext *tc, MVMObject *value) {
@@ -321,7 +344,7 @@ void * MVM_nativecall_unmarshal_cpointer(MVMThreadContext *tc, MVMObject *value)
         return ((MVMCPointer *)value)->body.ptr;
     else
         MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CPointer representation, but got a %s (%s)", REPR(value)->name, STABLE(value)->debug_name);
+            "Native call expected return type with CPointer representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
 }
 
 void * MVM_nativecall_unmarshal_carray(MVMThreadContext *tc, MVMObject *value) {
@@ -331,7 +354,7 @@ void * MVM_nativecall_unmarshal_carray(MVMThreadContext *tc, MVMObject *value) {
         return ((MVMCArray *)value)->body.storage;
     else
         MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CArray representation, but got a %s (%s)", REPR(value)->name, STABLE(value)->debug_name);
+            "Native call expected return type with CArray representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
 }
 
 void * MVM_nativecall_unmarshal_vmarray(MVMThreadContext *tc, MVMObject *value) {
@@ -345,7 +368,7 @@ void * MVM_nativecall_unmarshal_vmarray(MVMThreadContext *tc, MVMObject *value)
     }
     else
         MVM_exception_throw_adhoc(tc,
-            "Native call expected object with Array representation, but got a %s (%s)", REPR(value)->name, STABLE(value)->debug_name);
+            "Native call expected object with Array representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
 }
 
 void * MVM_nativecall_unmarshal_cunion(MVMThreadContext *tc, MVMObject *value) {
@@ -355,7 +378,7 @@ void * MVM_nativecall_unmarshal_cunion(MVMThreadContext *tc, MVMObject *value) {
         return ((MVMCUnion *)value)->body.cunion;
     else
         MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CUnion representation, but got a %s (%s)", REPR(value)->name, STABLE(value)->debug_name);
+            "Native call expected return type with CUnion representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
 }
 
 #ifdef _WIN32
@@ -365,13 +388,240 @@ static const char *dlerror(void)
     DWORD dw = GetLastError();
     if (dw == 0)
         return NULL;
-    sprintf(buf, "error 0x%x", (unsigned int)dw);
+    snprintf(buf, 32, "error 0x%"PRIx32"", (MVMuint32)dw);
     return buf;
 }
 #endif
 
+void init_c_call_node(MVMThreadContext *tc, MVMSpeshGraph *sg, MVMJitNode *node, void *func_ptr, MVMint16 num_args, MVMJitCallArg *args) {
+    node->type = MVM_JIT_NODE_CALL_C;
+    node->u.call.func_ptr = func_ptr;
+    if (0 < num_args) {
+        node->u.call.args = MVM_spesh_alloc(tc, sg, num_args * sizeof(MVMJitCallArg));
+        memcpy(node->u.call.args, args, num_args * sizeof(MVMJitCallArg));
+    }
+    else {
+        node->u.call.args = NULL;
+    }
+    node->u.call.num_args = num_args;
+    node->u.call.has_vargs = 0;
+    node->u.call.rv_mode = MVM_JIT_RV_VOID;
+    node->u.call.rv_idx = -1;
+}
+
+void init_box_call_node(MVMThreadContext *tc, MVMSpeshGraph *sg, MVMJitNode *box_rv_node, void *func_ptr, MVMint16 restype, MVMint16 dst) {
+    MVMJitCallArg args[] = { { MVM_JIT_INTERP_VAR , { MVM_JIT_INTERP_TC } },
+                             { MVM_JIT_REG_DYNIDX, { 2 } },
+                             { MVM_JIT_SAVED_RV, { 0 } }};
+    init_c_call_node(tc, sg, box_rv_node, func_ptr, 3, args);
+    box_rv_node->next = NULL;
+    if (dst == -1) {
+        box_rv_node->u.call.rv_mode = MVM_JIT_RV_DYNIDX;
+        box_rv_node->u.call.rv_idx = 0;
+    }
+    else {
+        box_rv_node->u.call.args[1].type = MVM_JIT_REG_VAL;
+        box_rv_node->u.call.args[1].v.reg = restype;
+
+        box_rv_node->u.call.rv_mode = MVM_JIT_RV_PTR;
+        box_rv_node->u.call.rv_idx = dst;
+    }
+}
+
+MVMJitGraph *MVM_nativecall_jit_graph_for_caller_code(
+    MVMThreadContext   *tc,
+    MVMSpeshGraph      *sg,
+    MVMNativeCallBody  *body,
+    MVMint16            restype,
+    MVMint16            dst,
+    MVMSpeshIns       **arg_ins
+) {
+    MVMJitGraph *jg = MVM_spesh_alloc(tc, sg, sizeof(MVMJitGraph)); /* will actually calloc */
+    MVMJitNode *block_gc_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
+    MVMJitNode *unblock_gc_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
+    MVMJitNode *call_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
+    MVMJitNode *save_rv_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
+    MVMJitNode *box_rv_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
+    MVMJitCode *jitcode = NULL;
+    MVMJitCallArg block_gc_args[] = { { MVM_JIT_INTERP_VAR , { MVM_JIT_INTERP_TC } } };
+
+    jg->sg = sg;
+    jg->first_node = block_gc_node;
+
+    save_rv_node->type = MVM_JIT_NODE_SAVE_RV;
+    save_rv_node->u.stack.slot = 0;
+    save_rv_node->next = unblock_gc_node;
+
+    init_c_call_node(tc, sg, block_gc_node,   &MVM_gc_mark_thread_blocked, 1, block_gc_args);
+    block_gc_node->next = call_node;
+
+    init_c_call_node(tc, sg, unblock_gc_node, &MVM_gc_mark_thread_unblocked, 1, block_gc_args);
+
+    init_c_call_node(tc, sg, call_node, body->entry_point, 0, NULL); /* we handle args manually */
+    call_node->next = save_rv_node;
+    call_node->u.call.num_args = body->num_args;
+    jg->last_node = unblock_gc_node->next = box_rv_node;
+
+    if (0 < body->num_args) {
+        MVMuint16 i = 0, str_arg_count = 0;
+        call_node->u.call.args = MVM_spesh_alloc(tc, sg, body->num_args * sizeof(MVMJitCallArg));
+        for (i = 0; i < body->num_args; i++) {
+            if ((body->arg_types[i] & MVM_NATIVECALL_ARG_TYPE_MASK) == MVM_NATIVECALL_ARG_UTF8STR) {
+                MVMJitNode *unbox_str_node;
+                MVMJitNode *save_str_rv_node;
+                MVMJitNode *free_str_node;
+
+                if (7 < ++str_arg_count) /* only got 7 empty slots in the stack scratch space */
+                    goto fail;
+
+                save_str_rv_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
+                save_str_rv_node->type = MVM_JIT_NODE_SAVE_RV;
+                save_str_rv_node->u.stack.slot = str_arg_count;
+                save_str_rv_node->next = jg->first_node;
+
+                unbox_str_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
+                {
+                    MVMJitCallArg unbox_str_args[] = {
+                        { MVM_JIT_INTERP_VAR , { MVM_JIT_INTERP_TC } },
+                        {
+                            dst == -1 ? MVM_JIT_ARG_I64 : MVM_JIT_PARAM_I64 ,
+                            { dst == -1 ? i : arg_ins[i]->operands[1].reg.orig }
+                        }
+                    };
+                    init_c_call_node(tc, sg, unbox_str_node, &MVM_string_utf8_maybe_encode_C_string, 2, unbox_str_args);
+                }
+                unbox_str_node->next = save_str_rv_node;
+                jg->first_node = unbox_str_node;
+
+                call_node->u.call.args[i].type = MVM_JIT_SAVED_RV;
+                call_node->u.call.args[i].v.lit_i64 = str_arg_count;
+
+                if ((body->arg_types[i] & MVM_NATIVECALL_ARG_FREE_STR_MASK) != 0) {
+                    MVMJitCallArg free_str_args[] = {
+                        { MVM_JIT_SAVED_RV , { str_arg_count } }
+                    };
+                    free_str_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
+                    init_c_call_node(tc, sg, free_str_node, &MVM_free, 1, free_str_args);
+                    free_str_node->next = unblock_gc_node->next;
+                    unblock_gc_node->next = free_str_node;
+                }
+            }
+        }
+        for (i = 0; i < body->num_args; i++) {
+            MVMJitArgType arg_type;
+            int is_rw = ((body->arg_types[i] & MVM_NATIVECALL_ARG_RW_MASK) == MVM_NATIVECALL_ARG_RW);
+
+            switch (body->arg_types[i] & MVM_NATIVECALL_ARG_TYPE_MASK) {
+                case MVM_NATIVECALL_ARG_CHAR:
+                case MVM_NATIVECALL_ARG_UCHAR:
+                case MVM_NATIVECALL_ARG_SHORT:
+                case MVM_NATIVECALL_ARG_USHORT:
+                case MVM_NATIVECALL_ARG_INT:
+                case MVM_NATIVECALL_ARG_UINT:
+                case MVM_NATIVECALL_ARG_LONG:
+                case MVM_NATIVECALL_ARG_ULONG:
+                case MVM_NATIVECALL_ARG_LONGLONG:
+                case MVM_NATIVECALL_ARG_ULONGLONG:
+                    arg_type = dst == -1
+                        ? is_rw ? MVM_JIT_ARG_I64_RW : MVM_JIT_ARG_I64
+                        : is_rw ? MVM_JIT_PARAM_I64_RW : MVM_JIT_PARAM_I64;
+                    break;
+                case MVM_NATIVECALL_ARG_CPOINTER:
+                    if (is_rw) goto fail;
+                    arg_type = dst == -1 ? MVM_JIT_ARG_PTR : MVM_JIT_PARAM_PTR;
+                    break;
+                case MVM_NATIVECALL_ARG_CARRAY:
+                    if (body->arg_types[i] & MVM_NATIVECALL_ARG_REFRESH_MASK) goto fail;
+                    if (is_rw) goto fail;
+                    arg_type = dst == -1 ? MVM_JIT_ARG_PTR : MVM_JIT_PARAM_PTR;
+                    break;
+                case MVM_NATIVECALL_ARG_VMARRAY:
+                    if (is_rw) goto fail;
+                    arg_type = dst == -1 ? MVM_JIT_ARG_VMARRAY : MVM_JIT_PARAM_VMARRAY;
+                    break;
+                case MVM_NATIVECALL_ARG_UTF8STR:
+                    if (is_rw) goto fail;
+                    continue; /* already handled */
+                default:
+                    goto fail;
+            }
+            call_node->u.call.args[i].type = arg_type;
+            call_node->u.call.args[i].v.lit_i64 = dst == -1 ? i : arg_ins[i]->operands[1].reg.orig;
+        }
+    }
+
+    if (body->ret_type == MVM_NATIVECALL_ARG_CHAR
+        || body->ret_type == MVM_NATIVECALL_ARG_UCHAR
+        || body->ret_type == MVM_NATIVECALL_ARG_SHORT
+        || body->ret_type == MVM_NATIVECALL_ARG_USHORT
+        || body->ret_type == MVM_NATIVECALL_ARG_INT
+        || body->ret_type == MVM_NATIVECALL_ARG_UINT
+        || body->ret_type == MVM_NATIVECALL_ARG_LONG
+        || body->ret_type == MVM_NATIVECALL_ARG_ULONG
+        || body->ret_type == MVM_NATIVECALL_ARG_LONGLONG
+        || body->ret_type == MVM_NATIVECALL_ARG_ULONGLONG
+    ) {
+        init_box_call_node(tc, sg, box_rv_node, &MVM_nativecall_make_int, restype, dst);
+    }
+    else if (body->ret_type == MVM_NATIVECALL_ARG_CPOINTER) {
+        init_box_call_node(tc, sg, box_rv_node, &MVM_nativecall_make_cpointer, restype, dst);
+    }
+    else if (body->ret_type == MVM_NATIVECALL_ARG_UTF8STR) {
+        MVMJitCallArg args[] = { { MVM_JIT_INTERP_VAR , { MVM_JIT_INTERP_TC } },
+                                 { MVM_JIT_REG_DYNIDX, { 2 } },
+                                 { MVM_JIT_LITERAL, { MVM_NATIVECALL_ARG_UTF8STR } },
+                                 { MVM_JIT_SAVED_RV, { 0 } }};
+        init_c_call_node(tc, sg, box_rv_node, &MVM_nativecall_make_str, 4, args);
+        box_rv_node->next = NULL;
+        if (dst == -1) {
+            box_rv_node->u.call.rv_mode = MVM_JIT_RV_DYNIDX;
+            box_rv_node->u.call.rv_idx = 0;
+        }
+        else {
+            box_rv_node->u.call.args[1].type = MVM_JIT_REG_VAL;
+            box_rv_node->u.call.args[1].v.reg = restype;
+
+            box_rv_node->u.call.rv_mode = MVM_JIT_RV_PTR;
+            box_rv_node->u.call.rv_idx = dst;
+        }
+    }
+    else if (body->ret_type == MVM_NATIVECALL_ARG_VOID) {
+        call_node->next = unblock_gc_node;
+        unblock_gc_node->next = NULL;
+        jg->last_node = unblock_gc_node;
+    }
+    else {
+        goto fail;
+    }
+
+    return jg;
+fail:
+    return NULL;
+}
+
+MVMJitCode *create_caller_code(MVMThreadContext *tc, MVMNativeCallBody *body) {
+    MVMSpeshGraph *sg = MVM_calloc(1, sizeof(MVMSpeshGraph));
+    MVMJitGraph *jg = MVM_nativecall_jit_graph_for_caller_code(tc, sg, body, -1, -1, NULL);
+    MVMJitCode *jitcode;
+    if (jg != NULL) {
+        MVMJitNode *entry_label = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
+        entry_label->next = jg->first_node;
+        jg->first_node = entry_label;
+        jg->num_labels = 1;
+
+        entry_label->type = MVM_JIT_NODE_LABEL;
+        entry_label->u.label.name = 0;
+        jitcode = MVM_jit_compile_graph(tc, jg);
+    }
+    else {
+        jitcode = NULL;
+    }
+    MVM_spesh_graph_destroy(tc, sg);
+    return jitcode;
+}
+
 /* Builds up a native call site out of the supplied arguments. */
-void MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
+MVMint8 MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
         MVMString *sym, MVMString *conv, MVMObject *arg_info, MVMObject *ret_info) {
     char *lib_name = MVM_string_utf8_c8_encode_C_string(tc, lib);
     char *sym_name = MVM_string_utf8_c8_encode_C_string(tc, sym);
@@ -388,7 +638,7 @@ void MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
 
     /* Try to load the library. */
     body->lib_name = lib_name;
-    body->lib_handle = MVM_nativecall_load_lib(strlen(lib_name) ? lib_name : NULL);
+    body->lib_handle = MVM_nativecall_load_lib(lib_name[0] ? lib_name : NULL);
 
     if (!body->lib_handle) {
         char *waste[] = { lib_name, NULL };
@@ -452,79 +702,26 @@ void MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
 #ifdef HAVE_LIBFFI
     body->ffi_ret_type = MVM_nativecall_get_ffi_type(tc, body->ret_type);
 #endif
+    if (tc->instance->jit_enabled) {
+        body->jitcode = create_caller_code(tc, body);
+    }
+    else
+        body->jitcode = NULL;
 
     MVM_telemetry_interval_stop(tc, interval_id, "nativecall built");
+
+    return body->jitcode != NULL;
 }
 
 static MVMObject * nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec, MVMObject *target_type, void *cpointer_body) {
     MVMObject *result = NULL;
 
-    MVMROOT(tc, target_spec, {
-        MVMROOT(tc, target_type, {
-            switch (REPR(target_type)->ID) {
-                case MVM_REPR_ID_P6opaque: {
-                    const MVMStorageSpec *ss = REPR(target_spec)->get_storage_spec(tc, STABLE(target_spec));
-                    if(ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_INT) {
-                        MVMint64 value = 0;
-                        if (ss->is_unsigned)
-                            switch(ss->bits) {
-                                case 8:
-                                    value = *(MVMuint8 *)cpointer_body;
-                                    break;
-                                case 16:
-                                    value = *(MVMuint16 *)cpointer_body;
-                                    break;
-                                case 32:
-                                    value = *(MVMuint32 *)cpointer_body;
-                                    break;
-                                case 64:
-                                default:
-                                    value = *(MVMuint64 *)cpointer_body;
-                            }
-                        else
-                            switch(ss->bits) {
-                                case 8:
-                                    value = *(MVMint8 *)cpointer_body;
-                                    break;
-                                case 16:
-                                    value = *(MVMint16 *)cpointer_body;
-                                    break;
-                                case 32:
-                                    value = *(MVMint32 *)cpointer_body;
-                                    break;
-                                case 64:
-                                default:
-                                    value = *(MVMint64 *)cpointer_body;
-                            }
-
-                        result = MVM_nativecall_make_int(tc, target_type, value);
-                    }
-                    else if(ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_NUM) {
-                        MVMnum64 value;
-
-                        switch(ss->bits) {
-                            case 32:
-                                value = *(MVMnum32 *)cpointer_body;
-                                break;
-                            case 64:
-                            default:
-                                value = *(MVMnum64 *)cpointer_body;
-                        }
-
-                        result = MVM_nativecall_make_num(tc, target_type, value);
-                    }
-                    else if(ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_STR) {
-                        result = MVM_nativecall_make_str(tc, target_type, MVM_NATIVECALL_ARG_UTF8STR,
-                        (char *)cpointer_body);
-                    }
-                    else
-                        MVM_exception_throw_adhoc(tc, "Internal error: unhandled target type");
-
-                    break;
-                }
-                case MVM_REPR_ID_P6int: {
-                    const MVMStorageSpec *ss = REPR(target_spec)->get_storage_spec(tc, STABLE(target_spec));
-                    MVMint64 value;
+    MVMROOT2(tc, target_spec, target_type, {
+        switch (REPR(target_type)->ID) {
+            case MVM_REPR_ID_P6opaque: {
+                const MVMStorageSpec *ss = REPR(target_spec)->get_storage_spec(tc, STABLE(target_spec));
+                if(ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_INT) {
+                    MVMint64 value = 0;
                     if (ss->is_unsigned)
                         switch(ss->bits) {
                             case 8:
@@ -555,11 +752,10 @@ static MVMObject * nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec,
                             default:
                                 value = *(MVMint64 *)cpointer_body;
                         }
+
                     result = MVM_nativecall_make_int(tc, target_type, value);
-                    break;
                 }
-                case MVM_REPR_ID_P6num: {
-                    const MVMStorageSpec *ss = REPR(target_spec)->get_storage_spec(tc, STABLE(target_spec));
+                else if(ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_NUM) {
                     MVMnum64 value;
 
                     switch(ss->bits) {
@@ -572,30 +768,89 @@ static MVMObject * nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec,
                     }
 
                     result = MVM_nativecall_make_num(tc, target_type, value);
-                    break;
                 }
-                case MVM_REPR_ID_MVMCStr:
-                case MVM_REPR_ID_P6str:
+                else if(ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_STR) {
                     result = MVM_nativecall_make_str(tc, target_type, MVM_NATIVECALL_ARG_UTF8STR,
-                        (char *)cpointer_body);
-                    break;
-                case MVM_REPR_ID_MVMCStruct:
-                    result = MVM_nativecall_make_cstruct(tc, target_type, (void *)cpointer_body);
-                    break;
-                case MVM_REPR_ID_MVMCUnion:
-                    result = MVM_nativecall_make_cunion(tc, target_type, (void *)cpointer_body);
-                    break;
-                case MVM_REPR_ID_MVMCPointer:
-                    result = MVM_nativecall_make_cpointer(tc, target_type, (void *)cpointer_body);
-                    break;
-                case MVM_REPR_ID_MVMCArray: {
-                    result = MVM_nativecall_make_carray(tc, target_type, (void *)cpointer_body);
-                    break;
+                    (char *)cpointer_body);
                 }
-                default:
+                else
                     MVM_exception_throw_adhoc(tc, "Internal error: unhandled target type");
+
+                break;
             }
-        });
+            case MVM_REPR_ID_P6int: {
+                const MVMStorageSpec *ss = REPR(target_spec)->get_storage_spec(tc, STABLE(target_spec));
+                MVMint64 value;
+                if (ss->is_unsigned)
+                    switch(ss->bits) {
+                        case 8:
+                            value = *(MVMuint8 *)cpointer_body;
+                            break;
+                        case 16:
+                            value = *(MVMuint16 *)cpointer_body;
+                            break;
+                        case 32:
+                            value = *(MVMuint32 *)cpointer_body;
+                            break;
+                        case 64:
+                        default:
+                            value = *(MVMuint64 *)cpointer_body;
+                    }
+                else
+                    switch(ss->bits) {
+                        case 8:
+                            value = *(MVMint8 *)cpointer_body;
+                            break;
+                        case 16:
+                            value = *(MVMint16 *)cpointer_body;
+                            break;
+                        case 32:
+                            value = *(MVMint32 *)cpointer_body;
+                            break;
+                        case 64:
+                        default:
+                            value = *(MVMint64 *)cpointer_body;
+                    }
+                result = MVM_nativecall_make_int(tc, target_type, value);
+                break;
+            }
+            case MVM_REPR_ID_P6num: {
+                const MVMStorageSpec *ss = REPR(target_spec)->get_storage_spec(tc, STABLE(target_spec));
+                MVMnum64 value;
+
+                switch(ss->bits) {
+                    case 32:
+                        value = *(MVMnum32 *)cpointer_body;
+                        break;
+                    case 64:
+                    default:
+                        value = *(MVMnum64 *)cpointer_body;
+                }
+
+                result = MVM_nativecall_make_num(tc, target_type, value);
+                break;
+            }
+            case MVM_REPR_ID_MVMCStr:
+            case MVM_REPR_ID_P6str:
+                result = MVM_nativecall_make_str(tc, target_type, MVM_NATIVECALL_ARG_UTF8STR,
+                    (char *)cpointer_body);
+                break;
+            case MVM_REPR_ID_MVMCStruct:
+                result = MVM_nativecall_make_cstruct(tc, target_type, (void *)cpointer_body);
+                break;
+            case MVM_REPR_ID_MVMCUnion:
+                result = MVM_nativecall_make_cunion(tc, target_type, (void *)cpointer_body);
+                break;
+            case MVM_REPR_ID_MVMCPointer:
+                result = MVM_nativecall_make_cpointer(tc, target_type, (void *)cpointer_body);
+                break;
+            case MVM_REPR_ID_MVMCArray: {
+                result = MVM_nativecall_make_carray(tc, target_type, (void *)cpointer_body);
+                break;
+            }
+            default:
+                MVM_exception_throw_adhoc(tc, "Internal error: unhandled target type");
+        }
     });
 
     return result;
@@ -609,7 +864,7 @@ MVMObject * MVM_nativecall_global(MVMThreadContext *tc, MVMString *lib, MVMStrin
     MVMObject *ret = NULL;
 
     /* Try to load the library. */
-    lib_handle = MVM_nativecall_load_lib(strlen(lib_name) ? lib_name : NULL);
+    lib_handle = MVM_nativecall_load_lib(lib_name[0] ? lib_name : NULL);
     if (!lib_handle) {
         char *waste[] = { lib_name, NULL };
         MVM_free(sym_name);
@@ -658,7 +913,7 @@ MVMObject * MVM_nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec, MV
         data_body = MVM_nativecall_unmarshal_vmarray(tc, source);
     else
         MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CPointer, CStruct, CArray, or VMArray representation, but got a %s (%s)", REPR(source)->name, STABLE(source)->debug_name);
+            "Native call expected return type with CPointer, CStruct, CArray, or VMArray representation, but got a %s (%s)", REPR(source)->name, MVM_6model_get_debug_name(tc, source));
     return nativecall_cast(tc, target_spec, target_type, data_body);
 }
 
@@ -681,7 +936,7 @@ MVMint64 MVM_nativecall_sizeof(MVMThreadContext *tc, MVMObject *obj) {
     else
         MVM_exception_throw_adhoc(tc,
             "NativeCall op sizeof expected type with CPointer, CStruct, CArray, P6int or P6num representation, but got a %s (%s)",
-            REPR(obj)->name, STABLE(obj)->debug_name);
+            REPR(obj)->name, MVM_6model_get_debug_name(tc, obj));
 }
 
 /* Write-barriers a dyncall object so that delayed changes to the C-side of
@@ -752,13 +1007,13 @@ void MVM_nativecall_refresh(MVMThreadContext *tc, MVMObject *cthingy) {
         for (i = 0; i < repr_data->num_attributes; i++) {
             MVMint32 kind = repr_data->attribute_locations[i] & MVM_CSTRUCT_ATTR_MASK;
             MVMint32 slot = repr_data->attribute_locations[i] >> MVM_CSTRUCT_ATTR_SHIFT;
-            void *cptr;   /* The pointer in the C storage. */
-            void *objptr; /* The pointer in the object representing the C object. */
+            void *cptr = NULL;   /* The pointer in the C storage. */
+            void *objptr = NULL; /* The pointer in the object representing the C object. */
 
             if (kind == MVM_CSTRUCT_ATTR_IN_STRUCT || !body->child_objs[slot])
                 continue;
 
-            cptr = *((void **)(storage + repr_data->struct_offsets[i]));
+            cptr = (void*)((uintptr_t)storage + (uintptr_t)repr_data->struct_offsets[i]);
             if (IS_CONCRETE(body->child_objs[slot])) {
                 switch (kind) {
                     case MVM_CSTRUCT_ATTR_CARRAY:
@@ -801,13 +1056,13 @@ void MVM_nativecall_refresh(MVMThreadContext *tc, MVMObject *cthingy) {
         for (i = 0; i < repr_data->num_attributes; i++) {
             MVMint32 kind = repr_data->attribute_locations[i] & MVM_CPPSTRUCT_ATTR_MASK;
             MVMint32 slot = repr_data->attribute_locations[i] >> MVM_CPPSTRUCT_ATTR_SHIFT;
-            void *cptr;   /* The pointer in the C storage. */
-            void *objptr; /* The pointer in the object representing the C object. */
+            void *cptr = NULL;   /* The pointer in the C storage. */
+            void *objptr = NULL; /* The pointer in the object representing the C object. */
 
             if (kind == MVM_CPPSTRUCT_ATTR_IN_STRUCT || !body->child_objs[slot])
                 continue;
 
-            cptr = *((void **)(storage + repr_data->struct_offsets[i]));
+            cptr = (void*)((uintptr_t)storage + (uintptr_t)repr_data->struct_offsets[i]);
             if (IS_CONCRETE(body->child_objs[slot])) {
                 switch (kind) {
                     case MVM_CPPSTRUCT_ATTR_CARRAY:
@@ -839,3 +1094,46 @@ void MVM_nativecall_refresh(MVMThreadContext *tc, MVMObject *cthingy) {
         }
     }
 }
+
+/* Locate the thread that a callback should be run on. */
+MVMThreadContext * MVM_nativecall_find_thread_context(MVMInstance *instance) {
+    MVMint64 wanted_thread_id = MVM_platform_thread_id();
+    MVMThreadContext *tc = NULL;
+    while (1) {
+        uv_mutex_lock(&(instance->mutex_threads));
+        if (instance->in_gc) {
+            /* VM is in GC; free lock since the GC will acquire it again to
+             * clear the in_gc flag, and sleep a bit until it's safe to read
+             * the threads list. */
+            uv_mutex_unlock(&(instance->mutex_threads));
+            MVM_platform_sleep(0.0001);
+        }
+        else {
+            /* Not in GC. If a GC starts while we are reading this, then we
+             * are holding mutex_threads, and the GC will block on it before
+             * it gets to a stage where it can move things. */
+            MVMThread *thread = instance->threads;
+            while (thread) {
+                if (thread->body.native_thread_id == wanted_thread_id) {
+                    tc = thread->body.tc;
+                    if (tc)
+                        break;
+                }
+                thread = thread->body.next;
+            }
+            if (!tc)
+                MVM_panic(1, "native callback ran on thread (%"PRId64") unknown to MoarVM",
+                    wanted_thread_id);
+            uv_mutex_unlock(&(instance->mutex_threads));
+            break;
+        }
+    }
+    return tc;
+}
+
+void MVM_nativecall_invoke_jit(MVMThreadContext *tc, MVMObject *site) {
+    MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, site);
+    MVMJitCode * const jitcode = body->jitcode;
+
+    jitcode->func_ptr(tc, *tc->interp_cu, jitcode->labels[0]);
+}
