@@ -27,9 +27,10 @@ MVMThreadContext * MVM_tc_create(MVMThreadContext *parent, MVMInstance *instance
 
     /* Set up GC nursery. We only allocate tospace initially, and allocate
      * fromspace the first time this thread GCs, provided it ever does. */
-    tc->nursery_tospace     = MVM_calloc(1, MVM_NURSERY_SIZE);
+    tc->nursery_tospace_size = MVM_gc_new_thread_nursery_size(instance);
+    tc->nursery_tospace     = MVM_calloc(1, tc->nursery_tospace_size);
     tc->nursery_alloc       = tc->nursery_tospace;
-    tc->nursery_alloc_limit = (char *)tc->nursery_alloc + MVM_NURSERY_SIZE;
+    tc->nursery_alloc_limit = (char *)tc->nursery_alloc + tc->nursery_tospace_size;
 
     /* Set up temporary root handling. */
     tc->num_temproots   = 0;
@@ -73,6 +74,9 @@ void MVM_tc_destroy(MVMThreadContext *tc) {
     /* We run once again (non-blocking) to eventually close filehandles. */
     uv_run(tc->loop, UV_RUN_NOWAIT);
 
+    /* Free specialization state. */
+    MVM_spesh_sim_stack_destroy(tc, tc->spesh_sim_stack);
+
     /* Free the nursery and finalization queue. */
     MVM_free(tc->nursery_fromspace);
     MVM_free(tc->nursery_tospace);
@@ -101,9 +105,6 @@ void MVM_tc_destroy(MVMThreadContext *tc) {
     MVM_free(tc->nfa_longlit);
     MVM_free(tc->multi_dim_indices);
 
-    /* Free per-thread lexotic cache. */
-    MVM_free(tc->lexotic_cache);
-
     /* Destroy the libuv event loop */
     uv_loop_delete(tc->loop);
 
