@@ -26,7 +26,9 @@ MVMObject * MVM_thread_new(MVMThreadContext *tc, MVMObject *invokee, MVMint64 ap
 
     /* Try to create the new threadcontext. Can throw if libuv can't
      * create a loop for it for some reason (i.e. too many open files) */
-    child_tc = MVM_tc_create(tc, tc->instance);
+    MVMROOT(tc, thread, {
+        child_tc = MVM_tc_create(tc, tc->instance);
+    });
 
     /* Set up the new threadcontext a little. */
     child_tc->thread_obj = thread;
@@ -52,13 +54,6 @@ static void thread_initial_invoke(MVMThreadContext *tc, void *data) {
     MVMObject *invokee = thread->body.invokee;
     thread->body.invokee = NULL;
 
-    /* Set up the cached current usecapture CallCapture (done here so
-     * we allocate it on the correct thread, and once the thread is
-     * active). */
-    MVMROOT(tc, invokee, {
-        tc->cur_usecapture = MVM_repr_alloc_init(tc, tc->instance->CallCapture);
-    });
-
     /* Create initial frame, which sets up all of the interpreter state also. */
     invokee = MVM_frame_find_invokee(tc, invokee, NULL);
     STABLE(invokee)->invoke(tc, invokee, MVM_callsite_get_common(tc, MVM_CALLSITE_ID_NULL_ARGS), NULL);
@@ -74,16 +69,25 @@ static void start_thread(void *data) {
     ThreadStart *ts = (ThreadStart *)data;
     MVMThreadContext *tc = ts->tc;
 
-    /* Stash thread ID. */
-    tc->thread_obj->body.native_thread_id = MVM_platform_thread_id();
-
     /* wait for the GC to finish if it's not finished stealing us. */
     MVM_gc_mark_thread_unblocked(tc);
     tc->thread_obj->body.stage = MVM_thread_stage_started;
 
+    /* Stash thread ID. */
+    tc->thread_obj->body.native_thread_id = MVM_platform_thread_id();
+
+    /* Create a spesh log for this thread, unless it's just going to run C
+     * code (and thus it's a VM internal worker). */
+    if (REPR(tc->thread_obj->body.invokee)->ID != MVM_REPR_ID_MVMCFunction)
+        MVM_spesh_log_initialize_thread(tc, 0);
+
+    MVM_debugserver_notify_thread_creation(tc);
+
     /* Enter the interpreter, to run code. */
     MVM_interp_run(tc, thread_initial_invoke, ts);
 
+    MVM_debugserver_notify_thread_destruction(tc);
+
     /* Pop the temp root stack's ts->thread_obj, if it's still there (if we
      * cleared the temp root stack on exception at some point, it'll already be
      * gone). */
@@ -105,33 +109,54 @@ static void start_thread(void *data) {
 /* Begins execution of a thread. */
 void MVM_thread_run(MVMThreadContext *tc, MVMObject *thread_obj) {
     MVMThread *child = (MVMThread *)thread_obj;
-    int status;
+    int status, added;
     ThreadStart *ts;
 
-    if (REPR(child)->ID == MVM_REPR_ID_MVMThread) {
-        MVMThread * volatile *threads;
+    if (REPR(child)->ID == MVM_REPR_ID_MVMThread && IS_CONCRETE(thread_obj)) {
         MVMThreadContext *child_tc = child->body.tc;
 
-        /* Move thread to starting stage. */
-        child->body.stage = MVM_thread_stage_starting;
-
-        /* Create thread state, to pass to the thread start callback. */
-        ts = MVM_malloc(sizeof(ThreadStart));
-        ts->tc = child_tc;
-        ts->thread_obj = thread_obj;
-
-        /* Push this to the *child* tc's temp roots. */
-        MVM_gc_root_temp_push(child_tc, (MVMCollectable **)&ts->thread_obj);
-
         /* Mark thread as GC blocked until the thread actually starts. */
         MVM_gc_mark_thread_blocked(child_tc);
 
-        /* Push to starting threads list */
-        threads = &tc->instance->threads;
-        do {
-            MVMThread *curr = *threads;
-            MVM_ASSIGN_REF(tc, &(child->common.header), child->body.next, curr);
-        } while (MVM_casptr(threads, child->body.next, child) != child->body.next);
+        /* Create thread state, to pass to the thread start callback. */
+        ts = MVM_malloc(sizeof(ThreadStart));
+        ts->tc = child_tc;
+
+        /* Push to starting threads list. We may need to retry this if we are
+         * asked to join a GC run at this point (since the GC would already
+         * have taken a snapshot of the thread list, so it's not safe to add
+         * another at this point). */
+        added = 0;
+        while (!added) {
+            uv_mutex_lock(&tc->instance->mutex_threads);
+            if (MVM_load(&tc->gc_status) == MVMGCStatus_NONE) {
+                /* Insert into list. */
+                MVM_ASSIGN_REF(tc, &(child->common.header), child->body.next,
+                    tc->instance->threads);
+                tc->instance->threads = child;
+
+                /* Store the thread object in the thread start information and
+                 * keep it alive by putting it in the *child* tc's temp roots. */
+                ts->thread_obj = thread_obj;
+                MVM_gc_root_temp_push(child_tc, (MVMCollectable **)&ts->thread_obj);
+
+                /* Move thread to starting stage. */
+                child->body.stage = MVM_thread_stage_starting;
+
+                /* Mark us done and unlock the mutex; any GC run will now have
+                 * a consistent view of the thread list and can safely run. */
+                added = 1;
+                uv_mutex_unlock(&tc->instance->mutex_threads);
+            }
+            else {
+                /* Another thread decided we'll GC now. Release mutex, and
+                 * do the GC, making sure thread_obj and child are marked. */
+                uv_mutex_unlock(&tc->instance->mutex_threads);
+                MVMROOT2(tc, thread_obj, child, {
+                    GC_SYNC_POINT(tc);
+                });
+            }
+        }
 
         /* Do the actual thread creation. */
         status = uv_thread_create(&child->body.thread, start_thread, ts);
@@ -169,7 +194,7 @@ static int try_join(MVMThreadContext *tc, MVMThread *thread) {
     return status;
 }
 void MVM_thread_join(MVMThreadContext *tc, MVMObject *thread_obj) {
-    if (REPR(thread_obj)->ID == MVM_REPR_ID_MVMThread) {
+    if (REPR(thread_obj)->ID == MVM_REPR_ID_MVMThread && IS_CONCRETE(thread_obj)) {
         int status = try_join(tc, (MVMThread *)thread_obj);
         if (status < 0)
             MVM_panic(MVM_exitcode_compunit, "Could not join thread: errorcode %d", status);
@@ -182,7 +207,7 @@ void MVM_thread_join(MVMThreadContext *tc, MVMObject *thread_obj) {
 
 /* Gets the (VM-level) ID of a thread. */
 MVMint64 MVM_thread_id(MVMThreadContext *tc, MVMObject *thread_obj) {
-    if (REPR(thread_obj)->ID == MVM_REPR_ID_MVMThread)
+    if (REPR(thread_obj)->ID == MVM_REPR_ID_MVMThread && IS_CONCRETE(thread_obj))
         return ((MVMThread *)thread_obj)->body.thread_id;
     else
         MVM_exception_throw_adhoc(tc,
@@ -192,7 +217,7 @@ MVMint64 MVM_thread_id(MVMThreadContext *tc, MVMObject *thread_obj) {
 /* Gets the native OS ID of a thread. If it's not yet available because
  * the thread was not yet started, this will return 0. */
 MVMint64 MVM_thread_native_id(MVMThreadContext *tc, MVMObject *thread_obj) {
-    if (REPR(thread_obj)->ID == MVM_REPR_ID_MVMThread)
+    if (REPR(thread_obj)->ID == MVM_REPR_ID_MVMThread && IS_CONCRETE(thread_obj))
         return ((MVMThread *)thread_obj)->body.native_thread_id;
     else
         MVM_exception_throw_adhoc(tc,
@@ -210,6 +235,18 @@ MVMObject * MVM_thread_current(MVMThreadContext *tc) {
     return (MVMObject *)tc->thread_obj;
 }
 
+/* Gets the number of locks held by a thread. */
+MVMint64 MVM_thread_lock_count(MVMThreadContext *tc, MVMObject *thread_obj) {
+    if (REPR(thread_obj)->ID == MVM_REPR_ID_MVMThread && IS_CONCRETE(thread_obj)) {
+        MVMThreadContext *thread_tc = ((MVMThread *)thread_obj)->body.tc;
+        return thread_tc ? thread_tc->num_locks : 0;
+    }
+    else {
+        MVM_exception_throw_adhoc(tc,
+            "Thread handle used with threadlockcount must have representation MVMThread");
+    }
+}
+
 void MVM_thread_cleanup_threads_list(MVMThreadContext *tc, MVMThread **head) {
     /* Assumed to be the only thread accessing the list.
      * must set next on every item. */
