@@ -35,8 +35,7 @@ static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
     MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, ri->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (nread >= 0) {
-        MVMROOT(tc, t, {
-        MVMROOT(tc, arr, {
+        MVMROOT2(tc, t, arr, {
             MVMArray *res_buf;
 
             /* Push the sequence number. */
@@ -55,18 +54,15 @@ static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
             /* Finally, no error. */
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
         });
-        });
     }
     else if (nread == UV_EOF) {
-        MVMROOT(tc, t, {
-        MVMROOT(tc, arr, {
+        MVMROOT2(tc, t, arr, {
             MVMObject *final = MVM_repr_box_int(tc,
                 tc->instance->boot_types.BOOTInt, ri->seq_number);
             MVM_repr_push_o(tc, arr, final);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
         });
-        });
         if (buf->base)
             MVM_free(buf->base);
         uv_read_stop(handle);
@@ -75,15 +71,13 @@ static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
     else {
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
-        MVMROOT(tc, t, {
-        MVMROOT(tc, arr, {
+        MVMROOT2(tc, t, arr, {
             MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
                 tc->instance->VMString, uv_strerror(nread));
             MVMObject *msg_box = MVM_repr_box_str(tc,
                 tc->instance->boot_types.BOOTStr, msg_str);
             MVM_repr_push_o(tc, arr, msg_box);
         });
-        });
         if (buf->base)
             MVM_free(buf->base);
         uv_read_stop(handle);
@@ -146,6 +140,17 @@ static void read_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_t
     }
 }
 
+/* Stops reading. */
+static void read_cancel(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
+    ReadInfo *ri = (ReadInfo *)data;
+    if (ri->work_idx >= 0) {
+        MVMIOAsyncSocketData *handle_data = (MVMIOAsyncSocketData *)ri->handle->body.data;
+        if (handle_data->handle && !uv_is_closing((uv_handle_t *)handle_data->handle))
+            uv_read_stop(handle_data->handle);
+        MVM_io_eventloop_remove_active_work(tc, &(ri->work_idx));
+    }
+}
+
 /* Marks objects for a read task. */
 static void read_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *worklist) {
     ReadInfo *ri = (ReadInfo *)data;
@@ -163,6 +168,7 @@ static void read_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
 static const MVMAsyncTaskOps read_op_table = {
     read_setup,
     NULL,
+    read_cancel,
     read_gc_mark,
     read_gc_free
 };
@@ -175,7 +181,8 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     /* Validate REPRs. */
     if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue)
         MVM_exception_throw_adhoc(tc,
-            "asyncreadbytes target queue must have ConcBlockingQueue REPR");
+            "asyncreadbytes target queue must have ConcBlockingQueue REPR (got %s)",
+             MVM_6model_get_stable_debug_name(tc, queue->st));
     if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
         MVM_exception_throw_adhoc(tc,
             "asyncreadbytes result type must have REPR AsyncTask");
@@ -189,15 +196,9 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     }
 
     /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-    MVMROOT(tc, buf_type, {
+    MVMROOT4(tc, queue, schedulee, h, buf_type, {
         task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
     });
-    });
-    });
-    });
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
     task->body.ops  = &read_op_table;
@@ -232,27 +233,23 @@ static void on_write(uv_write_t *req, int status) {
     MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, wi->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (status >= 0) {
-        MVMROOT(tc, arr, {
-        MVMROOT(tc, t, {
+        MVMROOT2(tc, arr, t, {
             MVMObject *bytes_box = MVM_repr_box_int(tc,
                 tc->instance->boot_types.BOOTInt,
                 wi->buf.len);
             MVM_repr_push_o(tc, arr, bytes_box);
         });
-        });
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
     }
     else {
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
-        MVMROOT(tc, arr, {
-        MVMROOT(tc, t, {
+        MVMROOT2(tc, arr, t, {
             MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
                 tc->instance->VMString, uv_strerror(status));
             MVMObject *msg_box = MVM_repr_box_str(tc,
                 tc->instance->boot_types.BOOTStr, msg_str);
             MVM_repr_push_o(tc, arr, msg_box);
         });
-        });
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
     MVM_free(wi->req);
@@ -343,6 +340,7 @@ static void write_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
 static const MVMAsyncTaskOps write_op_table = {
     write_setup,
     NULL,
+    NULL,
     write_gc_mark,
     write_gc_free
 };
@@ -366,15 +364,9 @@ static MVMAsyncTask * write_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObjec
         MVM_exception_throw_adhoc(tc, "asyncwritebytes requires a native array of uint8 or int8");
 
     /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-    MVMROOT(tc, buffer, {
+    MVMROOT4(tc, queue, schedulee, h, buffer, {
         task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
     });
-    });
-    });
-    });
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
     task->body.ops  = &write_op_table;
@@ -423,12 +415,12 @@ static void close_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
 static const MVMAsyncTaskOps close_op_table = {
     close_perform,
     NULL,
+    NULL,
     close_gc_mark,
     close_gc_free
 };
 
 static MVMint64 close_socket(MVMThreadContext *tc, MVMOSHandle *h) {
-    MVMIOAsyncSocketData *data = (MVMIOAsyncSocketData *)h->body.data;
     MVMAsyncTask *task;
     CloseInfo *ci;
 
@@ -462,9 +454,46 @@ static const MVMIOOps op_table = {
     NULL,
     NULL,
     NULL,
+    NULL,
     NULL
 };
 
+static void push_name_and_port(MVMThreadContext *tc, struct sockaddr_storage *name, MVMObject *arr) {
+    char addrstr[INET6_ADDRSTRLEN + 1];
+    /* XXX windows support kludge. 64 bit is much too big, but we'll
+     * get the proper data from the struct anyway, however windows
+     * decides to declare it. */
+    MVMuint64 port;
+    MVMObject *host_o;
+    MVMObject *port_o;
+    switch (name->ss_family) {
+        case AF_INET6: {
+            uv_ip6_name((struct sockaddr_in6*)name, addrstr, INET6_ADDRSTRLEN + 1);
+            port = ntohs(((struct sockaddr_in6*)name)->sin6_port);
+            break;
+        }
+        case AF_INET: {
+            uv_ip4_name((struct sockaddr_in*)name, addrstr, INET6_ADDRSTRLEN + 1);
+            port = ntohs(((struct sockaddr_in*)name)->sin_port);
+            break;
+        }
+        default:
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
+            return;
+            break;
+    }
+    MVMROOT(tc, arr, {
+        port_o = MVM_repr_box_int(tc, tc->instance->boot_types.BOOTInt, port);
+        MVMROOT(tc, port_o, {
+            host_o = (MVMObject *)MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr,
+                    MVM_string_ascii_decode_nt(tc, tc->instance->VMString, addrstr));
+        });
+    });
+    MVM_repr_push_o(tc, arr, host_o);
+    MVM_repr_push_o(tc, arr, port_o);
+}
+
 /* Info we convey about a connection attempt task. */
 typedef struct {
     struct sockaddr  *dest;
@@ -483,8 +512,7 @@ static void on_connect(uv_connect_t* req, int status) {
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (status >= 0) {
         /* Allocate and set up handle. */
-        MVMROOT(tc, arr, {
-        MVMROOT(tc, t, {
+        MVMROOT2(tc, arr, t, {
             MVMOSHandle          *result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
             MVMIOAsyncSocketData *data   = MVM_calloc(1, sizeof(MVMIOAsyncSocketData));
             data->handle                 = (uv_stream_t *)ci->socket;
@@ -492,19 +520,31 @@ static void on_connect(uv_connect_t* req, int status) {
             result->body.data            = data;
             MVM_repr_push_o(tc, arr, (MVMObject *)result);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
-        });
+
+            MVMROOT(tc, result, {
+                struct sockaddr_storage sockaddr;
+                int name_len = sizeof(struct sockaddr_storage);
+
+                uv_tcp_getpeername(ci->socket, (struct sockaddr *)&sockaddr, &name_len);
+                push_name_and_port(tc, &sockaddr, arr);
+
+                uv_tcp_getsockname(ci->socket, (struct sockaddr *)&sockaddr, &name_len);
+                push_name_and_port(tc, &sockaddr, arr);
+            });
         });
     }
     else {
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTIO);
-        MVMROOT(tc, arr, {
-        MVMROOT(tc, t, {
+        MVMROOT2(tc, arr, t, {
             MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
                 tc->instance->VMString, uv_strerror(status));
             MVMObject *msg_box = MVM_repr_box_str(tc,
                 tc->instance->boot_types.BOOTStr, msg_str);
             MVM_repr_push_o(tc, arr, msg_box);
-        });
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
         });
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
@@ -567,6 +607,7 @@ static const MVMAsyncTaskOps connect_op_table = {
     connect_setup,
     NULL,
     NULL,
+    NULL,
     connect_gc_free
 };
 
@@ -589,11 +630,9 @@ MVMObject * MVM_io_socket_connect_async(MVMThreadContext *tc, MVMObject *queue,
             "asyncconnect result type must have REPR AsyncTask");
 
     /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
+    MVMROOT2(tc, queue, schedulee, {
         task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
     });
-    });
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
     task->body.ops  = &connect_op_table;
@@ -618,6 +657,7 @@ typedef struct {
     int               backlog;
 } ListenInfo;
 
+
 /* Handles an incoming connection. */
 static void on_connection(uv_stream_t *server, int status) {
     ListenInfo       *li     = (ListenInfo *)server->data;
@@ -632,30 +672,42 @@ static void on_connection(uv_stream_t *server, int status) {
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if ((r = uv_accept(server, (uv_stream_t *)client)) == 0) {
         /* Allocate and set up handle. */
-        MVMROOT(tc, arr, {
-        MVMROOT(tc, t, {
+        MVMROOT2(tc, arr, t, {
             MVMOSHandle          *result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
             MVMIOAsyncSocketData *data   = MVM_calloc(1, sizeof(MVMIOAsyncSocketData));
             data->handle                 = (uv_stream_t *)client;
             result->body.ops             = &op_table;
             result->body.data            = data;
+
             MVM_repr_push_o(tc, arr, (MVMObject *)result);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
-        });
+
+            MVMROOT(tc, result, {
+                struct sockaddr_storage sockaddr;
+                int name_len = sizeof(struct sockaddr_storage);
+
+                uv_tcp_getpeername(client, (struct sockaddr *)&sockaddr, &name_len);
+                push_name_and_port(tc, &sockaddr, arr);
+
+                uv_tcp_getsockname(client, (struct sockaddr *)&sockaddr, &name_len);
+                push_name_and_port(tc, &sockaddr, arr);
+            });
         });
     }
     else {
         uv_close((uv_handle_t*)client, NULL);
         MVM_free(client);
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTIO);
-        MVMROOT(tc, arr, {
-        MVMROOT(tc, t, {
+        MVMROOT2(tc, arr, t, {
             MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
                 tc->instance->VMString, uv_strerror(r));
             MVMObject *msg_box = MVM_repr_box_str(tc,
                 tc->instance->boot_types.BOOTStr, msg_str);
             MVM_repr_push_o(tc, arr, msg_box);
-        });
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
         });
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
@@ -687,7 +739,12 @@ static void listen_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async
                     tc->instance->VMString, uv_strerror(r));
                 MVMObject *msg_box = MVM_repr_box_str(tc,
                     tc->instance->boot_types.BOOTStr, msg_str);
+
                 MVM_repr_push_o(tc, arr, msg_box);
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
             });
             MVM_repr_push_o(tc, t->body.queue, arr);
         });
@@ -708,7 +765,10 @@ static void on_listen_cancelled(uv_handle_t *handle) {
 }
 static void listen_cancel(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
     ListenInfo *li = (ListenInfo *)data;
-    uv_close((uv_handle_t *)li->socket, on_listen_cancelled);
+    if (li->socket) {
+        uv_close((uv_handle_t *)li->socket, on_listen_cancelled);
+        li->socket = NULL;
+    }
 }
 
 /* Frees info for a listen task. */
@@ -724,6 +784,7 @@ static void listen_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
 /* Operations table for async listen task. */
 static const MVMAsyncTaskOps listen_op_table = {
     listen_setup,
+    NULL,
     listen_cancel,
     NULL,
     listen_gc_free
@@ -748,11 +809,9 @@ MVMObject * MVM_io_socket_listen_async(MVMThreadContext *tc, MVMObject *queue,
             "asynclisten result type must have REPR AsyncTask");
 
     /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
+    MVMROOT2(tc, queue, schedulee, {
         task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
     });
-    });
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
     task->body.ops  = &listen_op_table;
