@@ -138,7 +138,7 @@ MVMString * MVM_dir_cwd(MVMThreadContext *tc) {
 #endif
 
     if ((r = uv_cwd(path, (size_t *)&max_path)) < 0) {
-        MVM_exception_throw_adhoc(tc, "chdir failed: %s", uv_strerror(r));
+        MVM_exception_throw_adhoc(tc, "Failed to determine cwd: %s", uv_strerror(r));
     }
 
     return MVM_string_utf8_c8_decode(tc, tc->instance->VMString, path, strlen(path));
@@ -198,6 +198,7 @@ static const MVMIOOps op_table = {
     NULL,
     NULL,
     NULL,
+    NULL,
     gc_free
 };
 
@@ -268,7 +269,7 @@ MVMObject * MVM_dir_open(MVMThreadContext *tc, MVMString *dirname) {
 static MVMOSHandle * get_dirhandle(MVMThreadContext *tc, MVMObject *oshandle, const char *msg) {
     MVMOSHandle *handle = (MVMOSHandle *)oshandle;
     if (REPR(oshandle)->ID != MVM_REPR_ID_MVMOSHandle)
-        MVM_exception_throw_adhoc(tc, "%s requires an object with REPR MVMOSHandle (got %s with REPR %s)", msg, STABLE(handle)->debug_name, REPR(handle)->name);
+        MVM_exception_throw_adhoc(tc, "%s requires an object with REPR MVMOSHandle (got %s with REPR %s)", msg, MVM_6model_get_debug_name(tc, (MVMObject *)handle), REPR(handle)->name);
     if (handle->body.ops != &op_table)
         MVM_exception_throw_adhoc(tc, "%s got incorrect kind of handle", msg);
     return handle;
