@@ -228,22 +228,20 @@ static MVMint32 MVM_jit_expr_add_const(MVMThreadContext *tc, MVMJitExprTree *tre
     return num;
 }
 
-static MVMint32 can_getlex(MVMThreadContext *tc, MVMJitGraph *jg, MVMSpeshIns *ins) {
-    MVMint32 outers = ins->operands[1].lex.outers;
-    MVMint32 idx    = ins->operands[1].lex.idx;
-    MVMStaticFrame *sf = jg->sg->sf;
-    MVMuint16 *lexical_types;
-    MVMint32 i;
-    for (i = 0; i < outers; i++) {
-        sf = sf->body.outer;
-    }
-    /* Use speshed lexical types, if necessary */
-    lexical_types = (outers == 0 && jg->sg->lexical_types != NULL ?
-                     jg->sg->lexical_types : sf->body.lexical_types);
-    /* can't do getlex yet, if we have an object register */
-    return lexical_types[idx] != MVM_reg_obj;
+static MVMint32 getlex_needs_autoviv(MVMThreadContext *tc, MVMJitGraph *jg, MVMSpeshIns *ins) {
+    MVMSpeshOperand opr = ins->operands[1];
+    MVMuint16 lexical_type = MVM_spesh_get_lex_type(tc, jg->sg, opr.lex.outers, opr.lex.idx);
+    return lexical_type == MVM_reg_obj;
 }
 
+static MVMint32 bindlex_needs_write_barrier(MVMThreadContext *tc, MVMJitGraph *jg, MVMSpeshIns *ins) {
+    MVMSpeshOperand opr = ins->operands[0];
+    MVMuint16 lexical_type = MVM_spesh_get_lex_type(tc, jg->sg, opr.lex.outers, opr.lex.idx);
+    /* need to hit a write barrier if we bindlex to a string */
+    return lexical_type == MVM_reg_obj || lexical_type == MVM_reg_str;
+}
+
+
 static MVMint32 ins_has_single_input_output_operand(MVMSpeshIns *ins) {
     switch (ins->info->opcode) {
     case MVM_OP_inc_i:
@@ -685,6 +683,9 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
     MVM_VECTOR_INIT(tree->info,  256);
     MVM_VECTOR_INIT(tree->roots, 16);
 
+    /* ensure that all references are nonzero */
+    MVM_VECTOR_PUSH(tree->nodes, MVM_JIT_NOOP);
+
     tree->graph      = jg;
     tree->num_labels = 0;
     /* Hold indices to the node that last computed a value belonging
@@ -714,7 +715,8 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
         struct ValueDefinition *defined_value = NULL;
 
         /* check if this is a getlex and if we can handle it */
-        BAIL(opcode == MVM_OP_getlex && !can_getlex(tc, jg, ins), "Can't compile object getlex");
+        BAIL(opcode == MVM_OP_getlex && getlex_needs_autoviv(tc, jg, ins), "Can't compile object getlex");
+        BAIL(opcode == MVM_OP_bindlex && bindlex_needs_write_barrier(tc, jg, ins), "Can't compile write-barrier bindlex");
 
         /* Check annotations that may require handling or wrapping the expression */
         for (ann = ins->annotations; ann != NULL; ann = ann->next) {
@@ -778,7 +780,14 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
             case MVM_SPESH_ANN_DEOPT_ONE_INS:
                 /* we should only see this in guards, which we don't do just
                  * yet, although we will. At the very least, this implies a flush. */
-                BAIL(1, "Cannot handle DEOPT_ONE (ins=%s)\n", ins->info->name);
+                switch (opcode) {
+                case MVM_OP_sp_guard:
+                case MVM_OP_sp_guardconc:
+                case MVM_OP_sp_guardtype:
+                case MVM_OP_sp_guardsf:
+                    BAIL(1, "Cannot handle DEOPT_ONE (ins=%s)\n", ins->info->name);
+                    break;
+                }
                 break;
             case MVM_SPESH_ANN_DEOPT_ALL_INS:
                 /* don't expect to be handling these, either, but these also
@@ -930,11 +939,13 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
     }
 
  done:
-    if (tree->nodes_num > 0) {
+    if (tree->roots_num > 0) {
         active_values_flush(tc, tree, values, sg->num_locals);
         MVM_jit_expr_tree_analyze(tc, tree);
         MVM_jit_log(tc, "Build tree out of: [");
         for (ins = entry; ins != iter->ins; ins = ins->next) {
+            if (ins->info->opcode == MVM_SSA_PHI)
+                continue;
             MVM_jit_log(tc, "%s, ", ins->info->name);
         }
         MVM_jit_log(tc, "]\n");
