@@ -5,14 +5,14 @@
 
 #if MVM_TRACING
 #  define TRACING_OPT "[--tracing] "
-#  define TRACING_USAGE "\n    --tracing  output a line to stderr on every interpreter instr"
+#  define TRACING_USAGE "\n    --tracing         output a line to stderr on every interpreter instr"
 #else
 #  define TRACING_OPT ""
 #  define TRACING_USAGE ""
 #endif
 
 #ifdef HAVE_TELEMEH
-#  define TELEMEH_USAGE "MVM_TELEMETRY_LOG           Log internal events at high precision to this file\n"
+#  define TELEMEH_USAGE "    MVM_TELEMETRY_LOG           Log internal events at high precision to this file\n"
 #else
 #  define TELEMEH_USAGE ""
 #endif
@@ -31,6 +31,10 @@
 #  define snprintf _snprintf
 #endif
 
+#if defined(_MSC_VER)
+#define strtoll _strtoi64
+#endif
+
 /* flags need to be sorted alphabetically */
 
 enum {
@@ -38,6 +42,7 @@ enum {
     UNKNOWN_FLAG = -1,
 
     FLAG_CRASH,
+    FLAG_SUSPEND,
     FLAG_DUMP,
     FLAG_FULL_CLEANUP,
     FLAG_HELP,
@@ -45,11 +50,13 @@ enum {
     FLAG_VERSION,
 
     OPT_EXECNAME,
-    OPT_LIBPATH
+    OPT_LIBPATH,
+    OPT_DEBUGPORT
 };
 
 static const char *const FLAGS[] = {
     "--crash",
+    "--debug-suspend",
     "--dump",
     "--full-cleanup",
     "--help",
@@ -67,23 +74,29 @@ USAGE: moar [--crash] [--libpath=...] " TRACING_OPT "input.moarvm [program args]
     --full-cleanup    try to free all memory and exit cleanly\n\
     --crash           abort instead of exiting on unhandled exception\n\
     --libpath         specify path loadbytecode should search in\n\
-    --version         show version information"
+    --version         show version information\n\
+    --debug-port=1234 listen for incoming debugger connections\n\
+    --debug-suspend   pause execution at the entry point"
     TRACING_USAGE
     "\n\
 \n\
 The following environment variables are respected:\n\
 \n\
     MVM_SPESH_DISABLE           Disables all dynamic optimization\n\
-    MVM_SPESH_NODELAY           Run dynamic optimization even for cold frames\n\
     MVM_SPESH_INLINE_DISABLE    Disables inlining\n\
     MVM_SPESH_OSR_DISABLE       Disables on-stack replacement\n\
+    MVM_SPESH_BLOCKING          Blocks log-sending thread while specializer runs\n\
+    MVM_SPESH_LOG               Specifies a dynamic optimizer log file\n\
+    MVM_SPESH_NODELAY           Run dynamic optimization even for cold frames\n\
     MVM_SPESH_LIMIT             Limit the maximum number of specializations\n\
     MVM_JIT_DISABLE             Disables JITting to machine code\n\
-    MVM_SPESH_LOG               Specifies a dynamic optimizer log file\n\
+    MVM_JIT_EXPR_DISABLE        Disable advanced 'expression' JIT\n\
     MVM_JIT_LOG                 Specifies a JIT-compiler log file\n\
     MVM_JIT_BYTECODE_DIR        Specifies a directory for JIT bytecode dumps\n\
     MVM_CROSS_THREAD_WRITE_LOG  Log unprotected cross-thread object writes to stderr\n\
-    MVM_COVERAGE_LOG            Append line-by-line coverage messages to this file\n"
+    MVM_COVERAGE_LOG            Append (de-duped by default) line-by-line coverage messages to this file\n\
+    MVM_COVERAGE_CONTROL        If set to 1, non-de-duping coverage started with nqp::coveragecontrol(1),\n\
+                                  if set to 2, non-de-duping coverage started right away\n"
     TELEMEH_USAGE;
 
 static int cmp_flag(const void *key, const void *value)
@@ -114,6 +127,8 @@ static int parse_flag(const char *arg)
         return OPT_LIBPATH;
     else if (starts_with(arg, "--execname="))
         return OPT_EXECNAME;
+    else if (starts_with(arg, "--debug-port="))
+        return OPT_DEBUGPORT;
     else
         return UNKNOWN_FLAG;
 }
@@ -140,6 +155,10 @@ int wmain(int argc, wchar_t *wargv[])
     int flag;
 
     unsigned int interval_id;
+    char telemeh_inited = 0;
+
+    MVMuint32 debugserverport = 0;
+    int start_suspended = 0;
 
     for (; (flag = parse_flag(argv[argi])) != NOT_A_FLAG; ++argi) {
         switch (flag) {
@@ -165,6 +184,10 @@ int wmain(int argc, wchar_t *wargv[])
             continue;
 #endif
 
+            case FLAG_SUSPEND:
+            start_suspended = 1;
+            continue;
+
             case OPT_EXECNAME:
             executable_name = argv[argi] + strlen("--execname=");
             continue;
@@ -198,6 +221,23 @@ int wmain(int argc, wchar_t *wargv[])
             return EXIT_SUCCESS;
             }
 
+            case OPT_DEBUGPORT: {
+                MVMint64 port;
+                char *portstr = argv[argi] + strlen("--debugport=") + 1;
+                char *endptr;
+                port = strtoll(portstr, &endptr, 10);
+                if (*endptr != '\0') {
+                    fprintf(stderr, "ERROR: Invalid characters in debug port flag: %s\n", portstr);
+                    return EXIT_FAILURE;
+                }
+                if (port <= 1024 || port > 65535) {
+                    fprintf(stderr, "ERROR: debug server port out of range. We only accept ports above 1024 and below 65535. (got: %lu)\n", port);
+                    return EXIT_FAILURE;
+                }
+                debugserverport = (MVMuint32)port;
+                break;
+            }
+
             default:
             fprintf(stderr, "ERROR: Unknown flag %s.\n\n%s\n", argv[argi], USAGE);
             return EXIT_FAILURE;
@@ -207,6 +247,7 @@ int wmain(int argc, wchar_t *wargv[])
 #ifdef HAVE_TELEMEH
     if (getenv("MVM_TELEMETRY_LOG")) {
         char path[256];
+        FILE *fp;
         snprintf(path, 255, "%s.%d", getenv("MVM_TELEMETRY_LOG"),
 #ifdef _WIN32
              _getpid()
@@ -214,8 +255,12 @@ int wmain(int argc, wchar_t *wargv[])
              getpid()
 #endif
              );
-        MVM_telemetry_init(fopen(path, "w"));
-        interval_id = MVM_telemetry_interval_start(0, "moarvm startup");
+        fp = fopen(path, "w");
+        if (fp) {
+            MVM_telemetry_init(fp);
+            telemeh_inited = 1;
+            interval_id = MVM_telemetry_interval_start(0, "moarvm startup");
+        }
     }
 #endif
 
@@ -242,10 +287,18 @@ int wmain(int argc, wchar_t *wargv[])
     signal(SIGPIPE, SIG_IGN);
 #endif
 
+    if (debugserverport > 0) {
+        MVM_debugserver_init(instance->main_thread, debugserverport);
+
+        if (start_suspended) {
+            instance->main_thread->gc_status = MVMGCStatus_INTERRUPT | MVMSuspendState_SUSPEND_REQUEST;
+        }
+    }
+
     if (dump) MVM_vm_dump_file(instance, input_file);
     else MVM_vm_run_file(instance, input_file);
 
-    if (getenv("MVM_TELEMETRY_LOG")) {
+    if (getenv("MVM_TELEMETRY_LOG") && telemeh_inited) {
         MVM_telemetry_interval_stop(0, interval_id, "moarvm teardown");
         MVM_telemetry_finish();
     }
