@@ -145,7 +145,7 @@ static mp_int * force_bigint(const MVMP6bigintBody *body, mp_int **tmp) {
         return body->u.bigint;
     }
     else {
-        MVMint32 value = body->u.smallint.value;
+        MVMint64 value = body->u.smallint.value;
         mp_int *i = MVM_malloc(sizeof(mp_int));
         mp_init(i);
         if (value >= 0) {
@@ -208,7 +208,7 @@ static void store_bigint_result(MVMP6bigintBody *body, mp_int *i) {
     }
 }
 
-/* Bitops on libtomath (no 2s compliment API) are horrendously inefficient and
+/* Bitops on libtomath (no two's complement API) are horrendously inefficient and
  * really should be hand-coded to work DIGIT-by-DIGIT with in-loop carry
  * handling.  For now we have these fixups.
  *
@@ -260,7 +260,7 @@ static void two_complement_bitop(mp_int *a, mp_int *b, mp_int *c,
         g = &e;
     }
     /* f and g now guaranteed to each point to positive bigints containing
-     * a 2s compliment representation of the values in a and b.  If either
+     * a two's complement representation of the values in a and b.  If either
      * a or b was negative, the representation is one tomath "digit" longer
      * than it need be and sign extended.
      */
@@ -332,11 +332,9 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     MVMObject *result; \
     mp_int *tmp[2] = { NULL, NULL }; \
     mp_int *ia, *ib, *ic; \
-    MVMROOT(tc, a, { \
-    MVMROOT(tc, b, { \
+    MVMROOT2(tc, a, b, { \
         result = MVM_repr_alloc_init(tc, result_type);\
     }); \
-    }); \
     ba = get_bigint_body(tc, a); \
     bb = get_bigint_body(tc, b); \
     bc = get_bigint_body(tc, result); \
@@ -360,11 +358,9 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
         mp_int *tmp[2] = { NULL, NULL }; \
         mp_int *ia, *ib, *ic; \
-        MVMROOT(tc, a, { \
-        MVMROOT(tc, b, { \
+        MVMROOT2(tc, a, b, { \
             result = MVM_repr_alloc_init(tc, result_type);\
         }); \
-        }); \
         ba = get_bigint_body(tc, a); \
         bb = get_bigint_body(tc, b); \
         bc = get_bigint_body(tc, result); \
@@ -398,11 +394,9 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     MVMP6bigintBody *bb = get_bigint_body(tc, b); \
     MVMP6bigintBody *bc; \
     MVMObject *result; \
-    MVMROOT(tc, a, { \
-    MVMROOT(tc, b, { \
+    MVMROOT2(tc, a, b, { \
         result = MVM_repr_alloc_init(tc, result_type);\
     }); \
-    }); \
     bc = get_bigint_body(tc, result); \
     if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
         mp_int *tmp[2] = { NULL, NULL }; \
@@ -442,11 +436,9 @@ MVMObject *MVM_bigint_gcd(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
     MVMP6bigintBody *bc;
     MVMObject       *result;
 
-    MVMROOT(tc, a, {
-    MVMROOT(tc, b, {
+    MVMROOT2(tc, a, b, {
         result = MVM_repr_alloc_init(tc, result_type);
     });
-    });
 
     bc = get_bigint_body(tc, result);
 
@@ -506,11 +498,9 @@ MVMObject * MVM_bigint_mod(MVMThreadContext *tc, MVMObject *result_type, MVMObje
 
     MVMObject *result;
 
-    MVMROOT(tc, a, {
-    MVMROOT(tc, b, {
+    MVMROOT2(tc, a, b, {
         result = MVM_repr_alloc_init(tc, result_type);
     });
-    });
 
     bc = get_bigint_body(tc, result);
 
@@ -554,11 +544,13 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
 
     int mp_result;
 
-    MVMROOT(tc, a, {
-    MVMROOT(tc, b, {
+    if (!MVM_BIGINT_IS_BIG(bb) && bb->u.smallint.value == 1 && STABLE(a) == STABLE(b)) {
+        return a;
+    }
+
+    MVMROOT2(tc, a, b, {
         result = MVM_repr_alloc_init(tc, result_type);
     });
-    });
 
     bc = get_bigint_body(tc, result);
 
@@ -672,7 +664,9 @@ MVMObject * MVM_bigint_pow(MVMThreadContext *tc, MVMObject *a, MVMObject *b,
             mp_int *ic = MVM_malloc(sizeof(mp_int));
             MVMP6bigintBody *resbody;
             mp_init(ic);
+            MVM_gc_mark_thread_blocked(tc);
             mp_expt_d(base, exponent_d, ic);
+            MVM_gc_mark_thread_unblocked(tc);
             r = MVM_repr_alloc_init(tc, int_type);
             resbody = get_bigint_body(tc, r);
             store_bigint_result(resbody, ic);
@@ -723,7 +717,7 @@ MVMObject *MVM_bigint_shl(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
  * a small int as well as cases when it is stored as a bigint */
 int BIGINT_IS_NEGATIVE (MVMP6bigintBody *ba) {
     mp_int *mp_a = ba->u.bigint;
-    if MVM_BIGINT_IS_BIG(ba) {
+    if (MVM_BIGINT_IS_BIG(ba)) {
         return SIGN(mp_a) == MP_NEG;
     }
     else {
@@ -827,6 +821,33 @@ void MVM_bigint_from_str(MVMThreadContext *tc, MVMObject *a, const char *buf) {
     }
 }
 
+MVMObject * MVM_bigint_from_bigint(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a) {
+    MVMP6bigintBody *a_body;
+    MVMP6bigintBody *r_body;
+    MVMObject       *result;
+
+    MVMROOT(tc, a, {
+        result = MVM_repr_alloc_init(tc, result_type);
+    });
+
+    a_body = get_bigint_body(tc, a);
+    r_body = get_bigint_body(tc, result);
+
+    if (MVM_BIGINT_IS_BIG(a_body)) {
+        mp_int *i = MVM_malloc(sizeof(mp_int));
+        mp_init_copy(i, a_body->u.bigint);
+        store_bigint_result(r_body, i);
+        adjust_nursery(tc, r_body);
+    }
+    else {
+        r_body->u.smallint       = a_body->u.smallint;
+        r_body->u.smallint.flag  = a_body->u.smallint.flag;
+        r_body->u.smallint.value = a_body->u.smallint.value;
+    }
+
+    return result;
+}
+
 MVMString * MVM_bigint_to_str(MVMThreadContext *tc, MVMObject *a, int base) {
     MVMP6bigintBody *body = get_bigint_body(tc, a);
     if (MVM_BIGINT_IS_BIG(body)) {
@@ -853,7 +874,7 @@ MVMString * MVM_bigint_to_str(MVMThreadContext *tc, MVMObject *a, int base) {
             char *buf;
             MVMString *result;
 
-            MVMint32 value = body->u.smallint.value;
+            MVMint64 value = body->u.smallint.value;
             mp_init(&i);
             if (value >= 0) {
                 mp_set_long(&i, value);
@@ -926,21 +947,65 @@ MVMnum64 MVM_bigint_div_num(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
     return c;
 }
 
-void MVM_bigint_rand(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b) {
+    MVMObject *result;
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb = get_bigint_body(tc, b);
 
-    mp_int *tmp[1] = { NULL };
-    mp_int *rnd = MVM_malloc(sizeof(mp_int));
-    mp_int *max = force_bigint(bb, tmp);
+    MVMint8 use_small_arithmetic = 0;
+    MVMint8 have_to_negate = 0;
+    MVMint32 smallint_max = 0;
 
-    mp_init(rnd);
-    mp_rand(rnd, USED(max) + 1);
+    if (MVM_BIGINT_IS_BIG(bb)) {
+        if (can_be_smallint(bb->u.bigint)) {
+            use_small_arithmetic = 1;
+            smallint_max = DIGIT(bb->u.bigint, 0);
+            have_to_negate = SIGN(bb->u.bigint) == MP_NEG;
+        }
+    } else {
+        use_small_arithmetic = 1;
+        smallint_max = bb->u.smallint.value;
+    }
 
-    mp_mod(rnd, max, rnd);
-    store_bigint_result(ba, rnd);
-    clear_temp_bigints(tmp, 1);
-    adjust_nursery(tc, ba);
+    if (use_small_arithmetic) {
+        if (MP_GEN_RANDOM_MAX >= abs(smallint_max)) {
+            mp_digit result_int = MP_GEN_RANDOM();
+            result_int = result_int % smallint_max;
+            if(have_to_negate)
+                result_int *= -1;
+
+            MVMROOT2(tc, type, b, {
+                result = MVM_repr_alloc_init(tc, type);
+            });
+
+            ba = get_bigint_body(tc, result);
+            store_int64_result(ba, result_int);
+        } else {
+            use_small_arithmetic = 0;
+        }
+    }
+
+    if (!use_small_arithmetic) {
+        mp_int *tmp[1] = { NULL };
+        mp_int *rnd = MVM_malloc(sizeof(mp_int));
+        mp_int *max = force_bigint(bb, tmp);
+
+        MVMROOT2(tc, type, b, {
+            result = MVM_repr_alloc_init(tc, type);
+        });
+
+        ba = get_bigint_body(tc, result);
+
+        mp_init(rnd);
+        mp_rand(rnd, USED(max) + 1);
+
+        mp_mod(rnd, max, rnd);
+        store_bigint_result(ba, rnd);
+        clear_temp_bigints(tmp, 1);
+        adjust_nursery(tc, ba);
+    }
+
+    return result;
 }
 
 MVMint64 MVM_bigint_is_prime(MVMThreadContext *tc, MVMObject *a, MVMint64 b) {
