@@ -27,12 +27,13 @@ void MVM_bigint_expmod(MVMThreadContext *tc, MVMObject *result, MVMObject *a, MV
 
 MVMint64 MVM_bigint_cmp(MVMThreadContext *tc, MVMObject *a, MVMObject *b);
 
+MVMObject * MVM_bigint_from_bigint(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a);
 void MVM_bigint_from_str(MVMThreadContext *tc, MVMObject *a, const char *buf);
 MVMString * MVM_bigint_to_str(MVMThreadContext *tc, MVMObject *a, int base);
 MVMnum64 MVM_bigint_to_num(MVMThreadContext *tc, MVMObject *a);
 MVMObject *MVM_bigint_from_num(MVMThreadContext *tc, MVMObject *result_type, MVMnum64 n);
 MVMnum64 MVM_bigint_div_num(MVMThreadContext *tc, MVMObject *a, MVMObject *b);
-void MVM_bigint_rand(MVMThreadContext *tc, MVMObject *a, MVMObject *b);
+MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *a, MVMObject *b);
 MVMint64 MVM_bigint_is_prime(MVMThreadContext *tc, MVMObject *a, MVMint64 b);
 MVMObject * MVM_bigint_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *str, MVMint64 offset, MVMint64 flag, MVMObject *type);
 MVMint64 MVM_bigint_is_big(MVMThreadContext *tc, MVMObject *a);
