@@ -199,7 +199,7 @@ MVMString * MVM_string_utf8_decode(MVMThreadContext *tc, const MVMObject *result
             MVMGrapheme32 g;
             ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, codepoint, &g);
             if (ready) {
-                while (count + ready >= bufsize) { /* if the buffer's full make a bigger one */
+                while (count + ready > bufsize) { /* if the buffer's full make a bigger one */
                     buffer = MVM_realloc(buffer, sizeof(MVMGrapheme32) * (
                         bufsize >= UTF8_MAXINC ? (bufsize += UTF8_MAXINC) : (bufsize *= 2)
                     ));
@@ -262,7 +262,7 @@ MVMString * MVM_string_utf8_decode(MVMThreadContext *tc, const MVMObject *result
     MVM_unicode_normalizer_eof(tc, &norm);
     ready = MVM_unicode_normalizer_available(tc, &norm);
     if (ready) {
-        if (count + ready >= bufsize) {
+        if (count + ready > bufsize) {
             buffer = MVM_realloc(buffer, sizeof(MVMGrapheme32) * (count + ready));
         }
         while (ready--) {
@@ -278,7 +278,7 @@ MVMString * MVM_string_utf8_decode(MVMThreadContext *tc, const MVMObject *result
     /* If we're lucky, we can fit our string in 8 bits per grapheme.
      * That happens when our lowest value is bigger than -129 and our
      * highest value is lower than 128. */
-    if (lowest_graph >= -128 && highest_graph < 128) {
+    if (-128 <= lowest_graph && highest_graph <= 127) {
         MVMGrapheme8 *new_buffer = MVM_malloc(sizeof(MVMGrapheme8) * count);
         for (ready = 0; ready < count; ready++) {
             new_buffer[ready] = buffer[ready];
@@ -325,10 +325,10 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
     MVMCodepoint codepoint = 0;
     MVMCodepoint lag_codepoint = -1;
     MVMint32 bufsize;
-    MVMGrapheme32 *buffer;
-    MVMDecodeStreamBytes *cur_bytes;
-    MVMDecodeStreamBytes *last_accept_bytes = ds->bytes_head;
-    MVMDecodeStreamBytes *lag_last_accept_bytes;
+    MVMGrapheme32 *buffer           = NULL;
+    MVMDecodeStreamBytes *cur_bytes = NULL;
+    MVMDecodeStreamBytes *last_accept_bytes     = ds->bytes_head;
+    MVMDecodeStreamBytes *lag_last_accept_bytes = NULL;
     MVMint32 last_accept_pos, lag_last_accept_pos, ready, at_start;
     MVMuint32 reached_stopper;
     MVMuint32 can_fast_path;
@@ -386,6 +386,8 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                 case UTF8_ACCEPT: {
                     if (codepoint == '\r' || codepoint >= first_significant) {
                         can_fast_path = 0;
+                        last_accept_bytes = cur_bytes;
+                        last_accept_pos = pos;
                         goto slow_path;
                     }
                     lag_codepoint = codepoint;
@@ -394,6 +396,7 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                     break;
                 }
                 case UTF8_REJECT:
+                    MVM_free(buffer);
                     MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
                     break;
                 }
@@ -409,6 +412,8 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                             &lag_codepoint, 1);
                         lag_codepoint = -1; /* Invalidate, we used it. */
                         can_fast_path = 0;
+                        last_accept_bytes = cur_bytes;
+                        last_accept_pos = pos;
                         goto slow_path;
                     }
 
@@ -425,13 +430,8 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                     }
                     buffer[count++] = lag_codepoint;
                     total++;
-                    if (MVM_string_decode_stream_maybe_sep(tc, seps, lag_codepoint)) {
-                        reached_stopper = 1;
-                        last_accept_bytes = lag_last_accept_bytes;
-                        last_accept_pos = lag_last_accept_pos;
-                        goto done;
-                    }
-                    else if (stopper_chars && *stopper_chars == total) {
+                    if (MVM_string_decode_stream_maybe_sep(tc, seps, lag_codepoint) ||
+                            stopper_chars && *stopper_chars == total) {
                         reached_stopper = 1;
                         last_accept_bytes = lag_last_accept_bytes;
                         last_accept_pos = lag_last_accept_pos;
@@ -445,12 +445,11 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                     break;
                 }
                 case UTF8_REJECT:
+                    MVM_free(buffer);
                     MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
                     break;
                 }
             }
-            lag_last_accept_bytes = cur_bytes;
-            lag_last_accept_pos = pos;
 
             /* If we fall out of the loop and have a lagged codepoint, but
              * no next buffer, then we fall into the slow path to process it
@@ -459,6 +458,8 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                 codepoint = lag_codepoint;
                 lag_codepoint = -1;
                 can_fast_path = 0;
+                last_accept_bytes = lag_last_accept_bytes;
+                last_accept_pos = lag_last_accept_pos;
                 goto slow_path;
             }
         }
@@ -468,10 +469,10 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                 case UTF8_ACCEPT: {
                     MVMGrapheme32 g;
                     MVMint32 first;
-                  slow_path:
-                    first = 1;
                     last_accept_bytes = cur_bytes;
                     last_accept_pos = pos;
+                  slow_path:
+                    first = 1;
                     ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc,
                         &(ds->norm), codepoint, &g);
                     while (ready--) {
@@ -489,11 +490,8 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                         }
                         buffer[count++] = g;
                         total++;
-                        if (MVM_string_decode_stream_maybe_sep(tc, seps, g)) {
-                            reached_stopper = 1;
-                            goto done;
-                        }
-                        else if (stopper_chars && *stopper_chars == total) {
+                        if (MVM_string_decode_stream_maybe_sep(tc, seps, g) ||
+                                stopper_chars && *stopper_chars == total) {
                             reached_stopper = 1;
                             goto done;
                         }
@@ -501,6 +499,7 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                     break;
                 }
                 case UTF8_REJECT:
+                    MVM_free(buffer);
                     MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
                     break;
                 }
@@ -527,10 +526,10 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
 char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
         MVMString *str, MVMuint64 *output_size, MVMint64 start, MVMint64 length,
         MVMString *replacement, MVMint32 translate_newlines) {
-    MVMuint8        *result;
+    MVMuint8        *result = NULL;
     size_t           result_pos, result_limit;
     MVMCodepointIter ci;
-    MVMStringIndex   strgraphs = MVM_string_graphs(tc, str);
+    MVMStringIndex   strgraphs  = MVM_string_graphs(tc, str);
     MVMuint8        *repl_bytes = NULL;
     MVMuint64        repl_length;
 
@@ -552,7 +551,7 @@ char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
     result_pos   = 0;
 
     /* Iterate the codepoints and encode them. */
-    MVM_string_ci_init(tc, &ci, str, translate_newlines);
+    MVM_string_ci_init(tc, &ci, str, translate_newlines, 0);
     while (MVM_string_ci_has_more(tc, &ci)) {
         MVMint32 bytes;
         MVMCodepoint cp = MVM_string_ci_get_codepoint(tc, &ci);
@@ -594,7 +593,7 @@ char * MVM_string_utf8_encode(MVMThreadContext *tc, MVMString *str, MVMuint64 *o
 /* Encodes the specified string to a UTF-8 C string. */
 char * MVM_string_utf8_encode_C_string(MVMThreadContext *tc, MVMString *str) {
     MVMuint64 output_size;
-    char * result;
+    char * result = NULL;
     char * utf8_string = MVM_string_utf8_encode(tc, str, &output_size, 0);
     /* this is almost always called from error-handling code. Don't care if it
      * contains embedded NULs. XXX TODO: Make sure all uses of this free what it returns */
@@ -605,6 +604,11 @@ char * MVM_string_utf8_encode_C_string(MVMThreadContext *tc, MVMString *str) {
     return result;
 }
 
+/* Encodes the specified string to a UTF-8 C string if it is not NULL. */
+char * MVM_string_utf8_maybe_encode_C_string(MVMThreadContext *tc, MVMString *str) {
+    return str ? MVM_string_utf8_encode_C_string(tc, str) : NULL;
+}
+
 void MVM_string_utf8_throw_encoding_exception (MVMThreadContext *tc, MVMCodepoint cp) {
     const char *gencat = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY);
     if(cp > 0x10FFFF) {
