@@ -213,12 +213,12 @@ static int append_grapheme(MVMThreadContext *tc, DecodeState *state, MVMGrapheme
     else if (g < 0) {
         MVMNFGSynthetic *synth = MVM_nfg_get_synthetic_info(tc, g);
         int mismatch = 0;
-        if (synth->base == state->orig_codes[state->orig_codes_unnormalized]) {
+        if (synth->codes[0] == state->orig_codes[state->orig_codes_unnormalized]) {
             MVMint32 i;
-            for (i = 0; i < synth->num_combs; i++) {
-                size_t orig_idx = state->orig_codes_unnormalized + i + 1;
-                if (orig_idx >= state->orig_codes_pos ||
-                        state->orig_codes[orig_idx] != synth->combs[i]) {
+            for (i = 1; i < synth->num_codes; i++) {
+                size_t orig_idx = state->orig_codes_unnormalized + i;
+                if (state->orig_codes_pos <= orig_idx ||
+                        state->orig_codes[orig_idx] != synth->codes[i]) {
                     mismatch = 1;
                     break;
                 }
@@ -229,7 +229,7 @@ static int append_grapheme(MVMThreadContext *tc, DecodeState *state, MVMGrapheme
         }
         if (!mismatch) {
             state->result[state->result_pos++] = g;
-            state->orig_codes_unnormalized += 1 + synth->num_combs;
+            state->orig_codes_unnormalized += synth->num_codes;
             return 1;
         }
     }
@@ -307,7 +307,12 @@ static void process_bad_bytes(MVMThreadContext *tc, DecodeState *state) {
         state->result[state->result_pos++] = synthetic_for(tc, state->utf8[i]);
     state->unaccepted_start = state->cur_byte + 1;
 }
-
+/* Check for if the codepoint is in range. Make sure it's not over 0x10FFFF
+ * and make sure it isn't a Surrogate */
+MVM_STATIC_INLINE int in_range (MVMCodepoint cp) {
+    return ( 0 <= cp && cp <= 0x10FFFF)
+        && (cp < 0xD800 || 0xDFFF < cp); /* Surrogates */
+}
 /* Decodes the specified number of bytes of utf8 into an NFG string, creating
  * a result of the specified type. The type must have the MVMString REPR. */
 MVMString * MVM_string_utf8_c8_decode(MVMThreadContext *tc, const MVMObject *result_type,
@@ -376,7 +381,7 @@ MVMString * MVM_string_utf8_c8_decode(MVMThreadContext *tc, const MVMObject *res
                                           | (decode_byte & 0x3F);
                     expected_continuations--;
                     if (expected_continuations == 0) {
-                        if (state.cur_codepoint >= min_expected_codepoint)
+                        if (min_expected_codepoint <= state.cur_codepoint && in_range(state.cur_codepoint))
                             process_ok_codepoint(tc, &state);
                         else
                             process_bad_bytes(tc, &state);
@@ -675,14 +680,13 @@ char * MVM_string_utf8_c8_encode_substr(MVMThreadContext *tc,
                     result_limit *= 2;
                     result = MVM_realloc(result, result_limit + 1);
                 }
-                result[result_pos++] = (hex2int(tc, synth->combs[1]) << 4) +
-                    hex2int(tc, synth->combs[2]);
+                result[result_pos++] = (hex2int(tc, synth->codes[2]) << 4) +
+                    hex2int(tc, synth->codes[3]);
             }
             else {
                 MVMint32 i;
-                emit_cp(tc, synth->base, &result, &result_pos, &result_limit, repl_bytes, repl_length);
-                for (i = 0; i < synth->num_combs; i++)
-                    emit_cp(tc, synth->combs[i], &result, &result_pos, &result_limit, repl_bytes, repl_length);
+                for (i = 0; i < synth->num_codes; i++)
+                    emit_cp(tc, synth->codes[i], &result, &result_pos, &result_limit, repl_bytes, repl_length);
             }
         }
     }
