@@ -273,3 +273,10 @@ typedef struct MVMHeapSnapshotReference MVMHeapSnapshotReference;
 typedef struct MVMHeapSnapshotState MVMHeapSnapshotState;
 typedef struct MVMHeapSnapshotWorkItem MVMHeapSnapshotWorkItem;
 typedef struct MVMHeapSnapshotSeen MVMHeapSnapshotSeen;
+typedef struct MVMDebugServerBreakpointInfo MVMDebugServerBreakpointInfo;
+typedef struct MVMDebugServerBreakpointFileTable MVMDebugServerBreakpointFileTable;
+typedef struct MVMDebugServerBreakpointTable MVMDebugServerBreakpointTable;
+typedef struct MVMDebugServerHandleTable MVMDebugServerHandleTable;
+typedef struct MVMDebugServerHandleTableEntry MVMDebugServerHandleTableEntry;
+typedef struct MVMDebugServerRequestEntry MVMDebugServerRequestEntry;
+typedef struct MVMDebugServerData MVMDebugServerData;
